from sklearn.tree._criterion cimport Criterion, RegressionCriterion
from sklearn.tree._tree cimport DTYPE_t         # Type of X
from sklearn.tree._tree cimport DOUBLE_t        # Type of y, sample_weight
from sklearn.tree._tree cimport SIZE_t          # Type for indices and counters
from ._bipartite_criterion cimport BipartiteCriterion, GMO
from ._axis_criterion cimport BaseComposableCriterion, AxisCriterion
from ._unsupervised_criterion cimport BaseUnsupervisedCriterion


cdef class BaseSemisupervisedAxisCriterion(AxisCriterion):
    """Abstract base class to facilitate future extension.

    This class should be used in type testing to allow future extensions
    to inherit from this class and still be accepted, not needing to subclass
    SSCompositeCriterion.
    """


cdef class SSCompositeCriterion(AxisCriterion):
    cdef public BaseComposableCriterion supervised_criterion
    cdef public BaseUnsupervisedCriterion unsupervised_criterion
    cdef const DOUBLE_t[:, ::1] X  # X data to initialize unuspervised criterion
    cdef SIZE_t n_features  # X.shape[1]

    # The supervision attribute is a float between 0 and 1 that weights
    # supervised and unsupervised impurities when calculating the total
    # semisupervised impurity:
    #
    #   final_impurity = (
    #     sup * self.supervised_criterion.node_impurity()
    #     / self._root_supervised_impurity
    #     + (1-sup) * self.unsupervised_criterion.node_impurity()
    #     / self._root_unsupervised_impurity
    #   )
    #
    # Its value can be dynamically controlled by a Python callable passed to
    # the constructor's update_supervision parameter, which should only receive
    # the current SSCompositeCriterion instance as argument and return a float.

    cdef double _root_supervised_impurity
    cdef double _root_unsupervised_impurity

    cdef double supervision           # first supervision value received
    cdef double _curr_supervision     # Current supervision amount
    cdef double _proxy_supervision    # Combined constants of proxy impurities
    cdef object update_supervision    # callable to update supervision

    cdef double _cached_u_impurity_left
    cdef double _cached_u_impurity_right
    cdef double _cached_s_impurity_left
    cdef double _cached_s_impurity_right

    # We test isinstance(self.supervised_criterion, AxisCriterion) only once at
    # __init__ to avoid requesting the GIL in methods that require that
    # condition.
    cdef bint _supervised_is_axis_criterion

    cpdef void set_X(self, const DOUBLE_t[:, ::1] X)
    cdef void set_root_impurities(self) noexcept nogil
    cdef void ss_children_impurities(
        self,
        double* u_impurity_left,
        double* u_impurity_right,
        double* s_impurity_left,
        double* s_impurity_right,
    ) noexcept nogil
    cdef inline void _copy_position_wise_attributes(self) noexcept nogil
    cdef int _update_supervision(self) except -1 nogil
    cdef int _set_proxy_supervision(self) except -1 nogil


cdef class BipartiteSemisupervisedCriterion(GMO):
    cdef public GMO bipartite_criterion
    cdef Criterion supervised_criterion_rows
    cdef Criterion supervised_criterion_cols
    cdef Criterion unsupervised_criterion_rows
    cdef Criterion unsupervised_criterion_cols

    cdef SIZE_t n_row_features
    cdef SIZE_t n_col_features

    cdef double _root_supervised_impurity
    cdef double _root_unsupervised_impurity_rows
    cdef double _root_unsupervised_impurity_cols

    cdef double _curr_supervision_rows
    cdef double _curr_supervision_cols

    cdef object update_supervision  # callable

    cpdef void set_X(
        self,
        const DOUBLE_t[:, ::1] X_rows,
        const DOUBLE_t[:, ::1] X_cols,
    )


# cython: boundscheck=False
from libc.stdint cimport SIZE_MAX

import numpy as np
from sklearn.tree._criterion cimport Criterion
from sklearn.tree._tree cimport SIZE_t

from ._bipartite_criterion import InvalidAxisError

cdef DOUBLE_t NAN = np.nan


cdef class SSCompositeCriterion(AxisCriterion):
    """Combines results from two criteria to yield its own.
    
    One criteria will receive y in its init() and the other will receive X.
    Their calculated impurities will then be combined as the final impurity:

        sup*supervised_impurity + (1-sup)*unsupervised_impurity

    where sup is self.supervision.

    When training with an unsupervised criterion, one must provide X and y
    stacked (joined cols) as the y parameter of the estimator's fit(). E.g.:

    >>> splitter = sklearn.tree._splitter.BestSplitter(*)
    >>> splitter.init(X=X, y=np.hstack([X, y]), *)
    >>> splitter.node_reset(*)

    Which will call

    >>> splitter.criterion.init(y=np.hstack([X, y], *)
    """
    def __reduce__(self):
        return (
            type(self),
            (
                self.supervised_criterion,
                self.unsupervised_criterion,
                self.supervision,
                self.update_supervision,
            ),
            self.__getstate__(),
        )

    def __getstate__(self):
        return {}

    def __cinit__(
        self,
        Criterion supervised_criterion,
        Criterion unsupervised_criterion,
        double supervision,
        object update_supervision=None,
    ):
        self.X = None
        self._cached_pos = SIZE_MAX
        self._root_supervised_impurity = 0.0
        self._root_unsupervised_impurity = 0.0
        self._curr_supervision = NAN

    def __init__(
        self,
        Criterion supervised_criterion,
        Criterion unsupervised_criterion,
        double supervision,
        object update_supervision=None,  # callable
    ):
        self.supervised_criterion = supervised_criterion
        self.unsupervised_criterion = unsupervised_criterion
        self.n_outputs = self.supervised_criterion.n_outputs
        self.n_samples = self.supervised_criterion.n_samples
        self.n_features = self.unsupervised_criterion.n_outputs

        self.supervision = supervision
        self._curr_supervision = supervision
        self.update_supervision = update_supervision
        self._supervised_is_axis_criterion = isinstance(
            self.supervised_criterion, AxisCriterion
        )
    
    cpdef void set_X(self, const DOUBLE_t[:, ::1] X):
        self.X = X

    cdef inline void _copy_position_wise_attributes(self) noexcept nogil:
        # NOTE: we assume the weighted_n_left/right of the supervised
        # criterion, that could sometimes diverge from the ones of the
        # unsupervised criterion.
        self.pos = self.supervised_criterion.pos
        self.weighted_n_left = self.supervised_criterion.weighted_n_left
        self.weighted_n_right = self.supervised_criterion.weighted_n_right   

    cdef int _update_supervision(self) except -1 nogil:
        # nogil to allow children classes to override it
        with gil:
            self._curr_supervision = self.update_supervision(
                y=self.y_,
                sample_indices=self.sample_indices,
                col_indices=self.col_indices,
                start=self.start,
                end=self.end,
                start_col=self.start_col,
                end_col=self.end_col,
                weighted_n_samples=self.weighted_n_samples,
                weighted_n_node_samples=self.weighted_n_node_samples,
                weighted_n_cols=self.weighted_n_cols,
                weighted_n_node_cols=self.weighted_n_node_cols,
                current_supervision=self._curr_supervision,
                original_supervision=self.supervision,
            )

    cdef void set_root_impurities(self) noexcept nogil:
        self._root_supervised_impurity = (
            self.supervised_criterion.node_impurity()
        )
        self._root_unsupervised_impurity = (
            self.unsupervised_criterion.node_impurity()
        )

    cdef int init(
        self, const DOUBLE_t[:, ::1] y,
        DOUBLE_t[:] sample_weight,
        double weighted_n_samples,
        SIZE_t[:] sample_indices,
        SIZE_t start,
        SIZE_t end,
    ) except -1 nogil:
        if self.X is None:
            with gil:
                raise RuntimeError("You must set_X() before init()")
        self._cached_pos = SIZE_MAX  # Important to reset cached values.
        self.y = y
        self.y_ = y  # to be used by self._update_supervision()
        self.sample_weight = sample_weight
        self.sample_indices = sample_indices
        self.start = start
        self.end = end
        self.n_node_samples = end - start
        self.weighted_n_samples = weighted_n_samples
        
        self.supervised_criterion.init(
            self.y,
            sample_weight,
            weighted_n_samples,
            sample_indices,
            start,
            end,
        )
        self.unsupervised_criterion.init(
            self.X,
            sample_weight,
            weighted_n_samples,
            sample_indices,
            start,
            end,
        )

        self.weighted_n_node_samples = \
            self.supervised_criterion.weighted_n_node_samples

        if self.update_supervision is not None:
            self._update_supervision()

        if self._root_supervised_impurity == 0.0:
            # Root unupervised impurities == 0.0 as well
            self.set_root_impurities()
        
        self._copy_position_wise_attributes()  # Because criteria were reset
        self._set_proxy_supervision()

        return 0

    cdef int axis_init(
        self,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
        const DOUBLE_t[:] col_weights,
        const SIZE_t[:] sample_indices,
        const SIZE_t[:] col_indices,
        double weighted_n_samples,
        double weighted_n_cols,
        SIZE_t start,
        SIZE_t end,
        SIZE_t start_col,
        SIZE_t end_col,
    ) except -1 nogil:
        if self.X is None:
            with gil:
                raise RuntimeError("You must set_X() before init()")
        if not self._supervised_is_axis_criterion:
            with gil:
                raise TypeError("supervised_criterion is not an AxisCriterion")

        self._cached_pos = SIZE_MAX  # Important to reset cached values.
        # Initialize fields
        self.y_ = y
        self.sample_weight = sample_weight
        self.sample_indices = sample_indices
        self.start = start
        self.end = end
        self.n_node_samples = end - start
        self.n_node_cols = end_col - start_col
        self.weighted_n_samples = weighted_n_samples

        # We do not call init_columns(), so we make these attributions here
        self.col_weights = col_weights
        self.weighted_n_cols = weighted_n_cols
        self.col_indices = col_indices
        self.start_col = start_col
        self.end_col = end_col

        (<AxisCriterion>self.supervised_criterion).axis_init(
            y,
            sample_weight=sample_weight,
            col_weights=col_weights,
            weighted_n_samples=weighted_n_samples,
            weighted_n_cols=weighted_n_cols,
            sample_indices=sample_indices,
            col_indices=col_indices,
            start=start,
            end=end,
            start_col=start_col,
            end_col=end_col,
        )

        self._columns_are_set = True  # Otherwise error would have been raised

        # We do not call self.init_columns(), but instead copy attributes from
        # the child criterion.
        self._node_col_indices = \
            (<AxisCriterion>self.supervised_criterion)._node_col_indices
        self.weighted_n_cols = \
            (<AxisCriterion>self.supervised_criterion).weighted_n_cols

        self.unsupervised_criterion.init(
            self.X,
            sample_weight,
            weighted_n_samples,
            sample_indices,
            start,
            end,
        )

        self.weighted_n_node_samples = \
            self.supervised_criterion.weighted_n_node_samples

        if self.update_supervision is not None:
            self._update_supervision()

        if self._root_supervised_impurity == 0.0:
            # Root unupervised impurities == 0.0 as well
            self.set_root_impurities()
        
        self._copy_position_wise_attributes()  # Because criteria were reset
        self._set_proxy_supervision()

        return 0

    cdef int reset(self) except -1 nogil:
        """Reset the criteria at pos=start."""
        self.supervised_criterion.reset()
        self.unsupervised_criterion.reset()
        self._copy_position_wise_attributes()
        return 0

    cdef int reverse_reset(self) except -1 nogil:
        """Reset the criteria at pos=end."""
        self.supervised_criterion.reverse_reset()
        self.unsupervised_criterion.reverse_reset()
        self._copy_position_wise_attributes()
        return 0

    cdef int update(self, SIZE_t new_pos) except -1 nogil:
        """Updated statistics by moving sample_indices[pos:new_pos] to the left child.
        This updates the collected statistics by moving sample_indices[pos:new_pos]
        from the right child to the left child.
        Parameters
        ----------
        new_pos : SIZE_t
            New starting index position of the sample_indices in the right child
        """
        self.supervised_criterion.update(new_pos)
        self.unsupervised_criterion.update(new_pos)
        self._copy_position_wise_attributes()
        return 0

    cdef double node_impurity(self) noexcept nogil:
        """Calculate the impurity of the node.
        Impurity of the current node, i.e. the impurity of sample_indices[start:end].
        This is the primary function of the criterion class. The smaller the
        impurity the better.
        """
        # FIXME: init is called before node_impurity by the TreeBuilder, where
        # root impurities are also set. We will redundantly calculate the root
        # impurity here as a consequence, divind it by itself to get 1 again.
        if self._root_supervised_impurity == 0.0:
            # Root unupervised impurities == 0.0 as well
            self.set_root_impurities()
            return 1.0

        cdef double sup = self._curr_supervision
        cdef double s_impurity = self.supervised_criterion.node_impurity()

        # If the node's Y partition is homogeneous (all values are equal), we
        # set the impurity to 0.0, disregarding the unsupervised criterion, to
        # avoid further redundant splitting (all descendant nodes would have the
        # same output value).
        if s_impurity == 0.0:
            return 0.0
        return (
            sup * s_impurity
            / self._root_supervised_impurity
            + (1.0-sup) * self.unsupervised_criterion.node_impurity()
            / self._root_unsupervised_impurity
        )

    cdef void ss_children_impurities(
        self,
        double* u_impurity_left,
        double* u_impurity_right,
        double* s_impurity_left,
        double* s_impurity_right,
    ) noexcept nogil:
        if self.pos == self._cached_pos:
            u_impurity_left[0] = self._cached_u_impurity_left 
            u_impurity_right[0] = self._cached_u_impurity_right
            s_impurity_left[0] = self._cached_s_impurity_left
            s_impurity_right[0] = self._cached_s_impurity_right
            return

        self.supervised_criterion.children_impurity(
            s_impurity_left, s_impurity_right,
        )
        self.unsupervised_criterion.children_impurity(
            u_impurity_left, u_impurity_right,
        )
        self._cached_pos = self.pos
        self._cached_u_impurity_left = u_impurity_left[0] 
        self._cached_u_impurity_right = u_impurity_right[0]
        self._cached_s_impurity_left = s_impurity_left[0]
        self._cached_s_impurity_right = s_impurity_right[0]

    cdef void children_impurity(
        self,
        double* impurity_left,
        double* impurity_right,
    ) noexcept nogil:
        """Calculate the impurity of children.
        Evaluate the impurity in children nodes, i.e. the impurity of
        sample_indices[start:pos] + the impurity of sample_indices[pos:end].

        Parameters
        ----------
        impurity_left : double pointer
            The memory address where the impurity of the left child should be
            stored.
        impurity_right : double pointer
            The memory address where the impurity of the right child should be
            stored
        """
        cdef:
            double u_impurity_left, u_impurity_right
            double s_impurity_left, s_impurity_right
            double sup = self._curr_supervision

        self.ss_children_impurities(
            &u_impurity_left,
            &u_impurity_right,
            &s_impurity_left,
            &s_impurity_right,
        )

        # If the node's Y partition is homogeneous (all values are equal), we
        # set the impurity to 0.0, disregarding the unsupervised criterion, to
        # avoid further redundant splitting (all descendant nodes would have the
        # same output value).
        if s_impurity_left == 0.0:
            impurity_left[0] = 0.0
        else:
            impurity_left[0] = (
                sup * s_impurity_left / self._root_supervised_impurity
                + (1.0 - sup) * u_impurity_left / self._root_unsupervised_impurity
            )
        if s_impurity_right == 0.0:
            impurity_right[0] = 0.0
        else:
            impurity_right[0] = (
                sup * s_impurity_right / self._root_supervised_impurity
                + (1.0 - sup) * u_impurity_right / self._root_unsupervised_impurity
            )

    cdef void node_value(self, double* dest) noexcept nogil:
        """Store the node value.
        Compute the node value of sample_indices[start:end] and save the value into
        dest.

        Parameters
        ----------
        dest : double pointer
            The memory address where the node value should be stored.
        """
        self.supervised_criterion.node_value(dest)

    cdef void total_node_value(self, double* dest) noexcept nogil:
        """Compute a single node value for all targets, disregarding y's shape.

        This method is used instead of node_value() in cases where the
        different columns of y are *not* considered as different outputs, being
        usually equivalent to node_value if y were to be flattened, i.e.

            total_node_value(y) == node_value(y.reshape[-1, 1])

        Parameters
        ----------
        dest : double pointer
            The memory address which we will save the node value into.
        """
        if self._supervised_is_axis_criterion:
            (<AxisCriterion>self.supervised_criterion).total_node_value(dest)
        else:
            with gil:
                raise TypeError(
                    "total value only available for supervised AxisCriterion"
                ) 

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
    ) noexcept nogil:
        """Compute the improvement in impurity.

        impurity_improvement = (
            self._curr_supervision
            * self.supervised_criterion.node_impurity()
            / self._root_supervised_impurity

            + (1 - self._curr_supervision)
            * self.unsupervised_criterion.node_impurity()
            / self._root_unsupervised_impurity
        )
        """
        # FIXME: Since we cannot access s_impurity and u_impurity
        # separately, we have to discard this method's input and calculate
        # them again. To mitigate this problem, self.ss_children_impurities()
        # caches the previous values it calculated and reuses them if
        # self.pos == self._cached_pos. If both criteria's impurity improvement
        # were the same linear combination of children impurities this would
        # not be a problem (as is the case with almost all sklearn Criterion
        # classes), but criteria such as FriedmanMSE would behave differently,
        # and we try to be as general as possible. Receiving a custom split
        # record instead of each impurity seems the ideal solution, since we
        # would be able to store and pass along the supervised and unsupervised
        # impurities separately.
        # NOTE: self._cached_pos must be reset at Criterion.init() to
        # ensure it always corresponds to the current tree node.
        cdef:
            double u_impurity_parent
            double s_impurity_parent
            double u_impurity_left, u_impurity_right
            double s_impurity_left, s_impurity_right
            double u_improvement
            double s_improvement
            double sup = self._curr_supervision

        self.ss_children_impurities(
            &u_impurity_left,
            &u_impurity_right,
            &s_impurity_left,
            &s_impurity_right,
        )

        u_impurity_parent = self.unsupervised_criterion.node_impurity()
        s_impurity_parent = self.supervised_criterion.node_impurity()

        # TODO: set impurity to 0.0 if s_impurity == 0.0? (See children_impurity)
        u_improvement = self.unsupervised_criterion.impurity_improvement(
            u_impurity_parent, u_impurity_left, u_impurity_right,
        ) / self._root_unsupervised_impurity

        s_improvement = self.supervised_criterion.impurity_improvement(
            s_impurity_parent, s_impurity_left, s_impurity_right,
        ) / self._root_supervised_impurity

        return sup * s_improvement + (1.0 - sup) * u_improvement

    cdef double proxy_impurity_improvement(self) noexcept nogil:
        """Compute a proxy of the impurity reduction.

        This method is used to speed up the search for the best split.
        It is a proxy quantity such that the split that maximizes this value
        also maximizes the impurity improvement. It neglects all constant terms
        of the impurity decrease for a given split.

        The absolute impurity improvement is only computed by the
        impurity_improvement method once the best split has been found.
        """
        return (
            self._proxy_supervision
            * self.supervised_criterion.proxy_impurity_improvement()
            + (1.0 - self._curr_supervision)  # here to avoid div by 0
            * self.unsupervised_criterion.proxy_impurity_improvement()
        )

    cdef int _set_proxy_supervision(self) except -1 nogil:
        """Combine all constants weighting the proxy impurities.

        Avoids performing these operations at every split position in
        the method proxy_impurity_improvement, performing them only once in
        init or axis_init.
        """
        self._proxy_supervision = (
            self._curr_supervision
            * self.unsupervised_criterion._proxy_improvement_factor()
            / self.supervised_criterion._proxy_improvement_factor()
            * self._root_unsupervised_impurity
            / self._root_supervised_impurity
        )

    cdef double _proxy_improvement_factor(self) noexcept nogil:
        """If improvement = proxy_improvement / a + b, this method returns a.

        This is useful when defining proxy impurity improvements for
        compositions of Criterion objects.
        """
        return (
            self.unsupervised_criterion._proxy_improvement_factor()
            * self._root_unsupervised_impurity
        )


# =============================================================================
# Bipartite Semi-supervised Criterion Wrapper
# =============================================================================


cdef class BipartiteSemisupervisedCriterion(GMO):

    def __cinit__(self, *args, **kwargs):
        self._root_supervised_impurity = 0.0
        self._root_unsupervised_impurity_rows = 0.0
        self._root_unsupervised_impurity_cols = 0.0

    def __init__(self, BipartiteCriterion bipartite_criterion):
        self.bipartite_criterion = bipartite_criterion
        self.criterion_rows = bipartite_criterion.criterion_rows
        self.criterion_cols = bipartite_criterion.criterion_cols

    cpdef void set_X(
        self,
        const DOUBLE_t[:, ::1] X_rows,
        const DOUBLE_t[:, ::1] X_cols,
    ):
        self.criterion_rows.set_X(X_rows)
        self.criterion_cols.set_X(X_cols)

    cdef int init(
        self,
        const DTYPE_t[:, :] X_rows,
        const DTYPE_t[:, :] X_cols,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] row_weights,
        const DOUBLE_t[:] col_weights,
        double weighted_n_rows,
        double weighted_n_cols,
        const SIZE_t[:] row_indices,
        const SIZE_t[:] col_indices,
        SIZE_t[2] start,
        SIZE_t[2] end,
    ) except -1 nogil:

        # Initialize fields
        self.X_rows = X_rows
        self.X_cols = X_cols
        self.y = y
        self.row_weights = row_weights
        self.col_weights = col_weights
        self.weighted_n_rows = weighted_n_rows
        self.weighted_n_cols = weighted_n_cols
        self.weighted_n_samples = weighted_n_rows * weighted_n_cols
        self.row_indices = row_indices
        self.col_indices = col_indices

        self.start[0] = start[0]
        self.start[1] = start[1]
        self.end[0] = end[0]
        self.end[1] = end[1]

        self.bipartite_criterion.init(
            self.X_rows,
            self.X_cols,
            self.y,
            self.row_weights,
            self.col_weights,
            self.weighted_n_rows,
            self.weighted_n_cols,
            self.row_indices,
            self.col_indices,
            self.start,
            self.end,
        )

        self.n_row_features = (<SSCompositeCriterion>self.criterion_rows).n_features
        self.n_col_features = (<SSCompositeCriterion>self.criterion_cols).n_features

        # Will be used by TreeBuilder as stopping criteria.
        self.weighted_n_node_rows = (
            self.bipartite_criterion.weighted_n_node_rows
        )
        self.weighted_n_node_cols = (
            self.bipartite_criterion.weighted_n_node_cols
        )

        # Will further be stored in the Tree object by the BipartiteSplitter.
        self.weighted_n_node_samples = (
            self.bipartite_criterion.weighted_n_node_samples
        )

        cdef double total_sup

        self._curr_supervision_rows = \
            (<SSCompositeCriterion>self.criterion_rows)._curr_supervision
        self._curr_supervision_cols = \
            (<SSCompositeCriterion>self.criterion_cols)._curr_supervision

        total_sup = (
            self._curr_supervision_rows + self._curr_supervision_cols
        )

        (<SSCompositeCriterion>self.criterion_rows)._curr_supervision = \
            total_sup / (self._curr_supervision_cols + 1)
        (<SSCompositeCriterion>self.criterion_cols)._curr_supervision = \
            total_sup / (self._curr_supervision_rows + 1)

        (<SSCompositeCriterion>self.criterion_rows)._set_proxy_supervision()
        (<SSCompositeCriterion>self.criterion_cols)._set_proxy_supervision()

    cdef void node_value(self, double* dest) noexcept nogil:
        self.bipartite_criterion.node_value(dest)

    cdef double node_impurity(self) noexcept nogil:
        return 0.5 * (
            self.criterion_rows.node_impurity()
            + self.criterion_cols.node_impurity()
        )

    cdef void children_impurity(
            self,
            double* impurity_left,
            double* impurity_right,
            SIZE_t axis,
    ) noexcept nogil:
        cdef:
            void* ss_criterion_ptr = self._get_criterion(axis)
            void* other_ss_criterion_ptr = self._get_criterion(1 - axis)
            void* u_other_criterion_ptr
            double other_sup
            double scaled_u_other_imp
            double u_other_root_imp

        u_other_criterion_ptr = <void*>(
            (<SSCompositeCriterion>other_ss_criterion_ptr)
            .unsupervised_criterion
        )
        u_other_root_imp = (
            (<SSCompositeCriterion>other_ss_criterion_ptr)
            ._root_unsupervised_impurity
        )
        if axis == 1:
            other_sup = self._curr_supervision_rows
        elif axis == 0:
            other_sup = self._curr_supervision_cols
        else:
            with gil:
                raise InvalidAxisError(axis)

        # There is no guarantee that the splitters are using the
        # unsupervised criteria, it is a valid option to use it only for
        # choosing an axis, calculating the unsupervised impurity only
        # here. Therefore, we must ensure the unsupervised criterion is in
        # the right position.
        # TODO: remove this dependency on the way we do axis_supervision_only.
        (<SSCompositeCriterion>ss_criterion_ptr).update(
            (<SSCompositeCriterion>ss_criterion_ptr).supervised_criterion.pos
        )

        scaled_u_other_imp = (
            (<Criterion>u_other_criterion_ptr).node_impurity()
            * (0.5 - 0.5 * other_sup)
            / u_other_root_imp
        )

        (<SSCompositeCriterion>ss_criterion_ptr).children_impurity(
            impurity_left,
            impurity_right,
        )

        impurity_left[0] = (
            (0.5 + 0.5 * other_sup) * impurity_left[0]
            + scaled_u_other_imp
        )
        impurity_right[0] = (
            (0.5 + 0.5 * other_sup) * impurity_right[0]
            + scaled_u_other_imp
        )

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
        SIZE_t axis,
    ) noexcept nogil:
        cdef void* criterion_ptr = self._get_criterion(axis)
        cdef double other_sup

        if axis == 0:
            other_sup = self._curr_supervision_cols
        elif axis == 1:
            other_sup = self._curr_supervision_rows
        else:
            with gil:
                raise InvalidAxisError(axis)

        # There is no guarantee that the splitters are using the
        # unsupervised criteria, it is a valid option to use it only for
        # choosing an axis, calculating the unsupervised impurity only
        # here. Therefore, we must ensure the unsupervised criterion is in
        # the right position.
        # TODO: remove this dependency on the way we do axis_supervision_only.
        (<SSCompositeCriterion>criterion_ptr).update(
            (<SSCompositeCriterion>criterion_ptr).supervised_criterion.pos
        )
        return (
            (<SSCompositeCriterion>criterion_ptr).impurity_improvement(
                impurity_parent, impurity_left, impurity_right
            )
            * (0.5 * other_sup + 0.5)
        )



from abc import ABCMeta, abstractmethod
from typing import Callable, Dict
import numpy as np
import pytest
import scipy.stats

from collections import defaultdict
from numbers import Real, Integral
from sklearn.metrics.pairwise import rbf_kernel
from sklearn.utils.validation import check_random_state
from sklearn.utils._param_validation import validate_params, Interval
from bipartite_learn.tree._splitter_factory import (
    make_semisupervised_criterion,
    make_bipartite_criterion,
    make_2dss_criterion,
)
from bipartite_learn.tree._axis_criterion import (
    AxisSquaredError,
    AxisSquaredErrorGSO,
    AxisFriedmanGSO,
    AxisGini,
    AxisEntropy,
)
from bipartite_learn.tree._unsupervised_criterion import (
    PairwiseFriedman,
    PairwiseSquaredError,
    PairwiseSquaredErrorGSO,
    UnsupervisedSquaredError,
    UnsupervisedGini,
    UnsupervisedEntropy,
    # UnsupervisedSquaredErrorGSO,  # TODO
    UnsupervisedFriedman,
    MeanDistance,
)
from bipartite_learn.tree._bipartite_criterion import GMO, GMOSA
from bipartite_learn.tree._semisupervised_criterion import (
    SSCompositeCriterion,
)

from bipartite_learn.melter import row_cartesian_product
from .utils.test_utils import assert_equal_dicts, comparison_text
from .utils.make_examples import make_interaction_blobs
from .utils.tree_utils import (
    apply_criterion,
    apply_ss_criterion,
    apply_bipartite_criterion,
    apply_bipartite_ss_criterion,
)

CLASSIFICATION_CRITERIA = {
    AxisGini,
    AxisEntropy,
    UnsupervisedGini,
    UnsupervisedEntropy,
}

PAIRWISE_CRITERIA = {
    MeanDistance,
}


# =============================================================================
# General fixtures
# =============================================================================


@pytest.fixture(params=[0.0, 0.00001, 0.1, 0.2328, 0.569, 0.782, 0.995, 1.0])
def supervision(request):
    return request.param


@pytest.fixture(params=range(10))
def random_state(request):
    return check_random_state(request.param)


@pytest.fixture
def n_samples(request):
    return (11, 7)


@pytest.fixture
def n_classes(n_samples):
    return (
        np.repeat(6, n_samples[1]),
        np.repeat(6, n_samples[0]),
    )


@pytest.fixture(params=[(3, 5), (2, 3)])
def n_features(request):
    return request.param


@pytest.fixture
def data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_blobs(
        return_molten=True,
        n_features=n_features,
        n_samples=n_samples,
        random_state=random_state,
        noise=2.0,
        centers=10,
    )
    X = [Xi.astype('float64') for Xi in X]
    return X, Y, x, y


# =============================================================================
# Criterion objects to compare against
# =============================================================================


class BaseReferenceCriterion(metaclass=ABCMeta):
    def __init__(self, average_both_axes=False):
        self.average_both_axes = average_both_axes

    def set_data(self, X, y, sample_weight=None, **kwargs):
        self.X = X
        self.y = y
        self.sample_weight = sample_weight

    @abstractmethod
    def impurity(self, y):
        ...

    def impurity_improvement(self, split_data: dict):
        return (
            split_data['weighted_n_node_samples'] /
            split_data['weighted_n_samples']
            * (
                split_data['impurity_parent']

                - split_data['impurity_left']
                * split_data['weighted_n_left']
                / split_data['weighted_n_node_samples']

                - split_data['impurity_right']
                * split_data['weighted_n_right']
                / split_data['weighted_n_node_samples']
            )
        )
    
    def children_impurity(self, y_region, rel_pos):
        return (
            self.impurity(y_region[:rel_pos]),
            self.impurity(y_region[rel_pos:]),
        )

    @validate_params(
        {
            'start': [Integral],
            'end': [Integral],
            'feature': [Interval(Integral, 0, None, closed='left'), None],
        },
        prefer_skip_nested_validation=False,  # Preferable for testing
    )
    def evaluate_split(
        self,
        pos: int,
        start: int = 0,
        end: int = 0,
        *,
        feature: int | None = None,
        weighted_n_samples: float | None = None,
        weighted_n_node_samples: float | None = None,
    ) -> Dict[str, int | float]:
        """Evaluate impurities and improvements of a split position.

        Calculate the impurities and impurity improvement of a given split
        position in a dataset, simulating split evaluation by the Criterion
        objects of decision trees.

        Parameters
        ----------
        pos : int, optional (default=0)
            The split position to evaluate. The absolute index must be provided,
            relative to the beginning of the dataset, NOT relative to `start`.
        start : int, optional (default=0)
            The starting position of the data to evaluate, representing the start
            index of the current node, by default 0.
        end : int, optional (default=0)
            The ending position of the data to evaluate, representing the final
            index of the current node, by default 0.
        feature : int, optional (default=None)
            The feature index representing the column of X to sort the data by. If
            None, no sorting is performed. By default None.
        weighted_n_samples : float, optional (default=None)
            The total weighted number of samples in the dataset, taken as
            `sample_weight.sum()` if None.
        weighted_n_node_samples : float, optional (default=None)
            The total weighted number of samples in the current node, taken to be
            `sample_weight[start:end].sum()` if None.

        Returns
        -------
        Dict[str, Number]
            A dictionary with at least the following keys and values:
            - 'pos': the split position.
            - 'impurity_parent': the impurity of the parent (current) node.
            - 'weighted_n_left': the weighted number of samples in the left node.
            - 'weighted_n_right': the weighted number of samples in the right node.
            - 'impurity_left': the impurity of the left node.
            - 'impurity_right': the impurity of the right node.
            - 'improvement': the impurity improvement gained by the split.
        """
        start = start if start >= 0 else self.y.shape[0] + start
        end = end if end > 0 else self.y.shape[0] + end

        if not (start < pos <= end):
            raise ValueError(
                f'Provided index {pos} is not between {start=} and {end=}.'
            )

        x_ = self.X[start:end].copy()
        y_ = self.y[start:end].copy()
        rel_pos = pos - start

        if feature is not None:
            x_, y_ = sort_by_feature(x_, y_, feature)

        n_samples = self.y.shape[0]
        n_node_samples = y_.shape[0]

        result = {
            'pos': pos,
            'impurity_parent': self.impurity(y_),
            'start': start,
            'end': end,
        }

        if self.sample_weight is None:
            result['weighted_n_left'] = float(rel_pos)
            weighted_n_samples = float(n_samples)
            weighted_n_node_samples = float(n_node_samples)
        else:
            result['weighted_n_left'] = self.sample_weight[start:pos].sum()
            if weighted_n_samples is None:
                weighted_n_samples = self.sample_weight.sum()
            if weighted_n_node_samples is None:
                weighted_n_node_samples = self.sample_weight[start:end].sum()

        result['weighted_n_right'] = (
            weighted_n_node_samples - result['weighted_n_left']
        )

        result['n_samples'] = n_samples
        result['n_node_samples'] = n_node_samples
        result['weighted_n_samples'] = weighted_n_samples
        result['weighted_n_node_samples'] = weighted_n_node_samples

        result['impurity_left'], result['impurity_right'] = (
            self.children_impurity(y_, rel_pos)
        )

        if self.average_both_axes:
            result['impurity_parent'] += self.impurity(y_.T)
            result['impurity_parent'] /= 2
            result['impurity_left'] += self.impurity(y_[:rel_pos].T)
            result['impurity_left'] /= 2
            result['impurity_right'] += self.impurity(y_[rel_pos:].T)
            result['impurity_right'] /= 2

        result['improvement'] = self.impurity_improvement(result)

        return result

    def eval_all_splits(
        self,
        start=0,
        end=0,
        indices=None,
    ):
        """Evaluate splits for all positions in a given range of indices.

        Parameters
        ----------
        start : int, optional (default=0)
            The starting index for the range of indices to evaluate splits for.
            If negative, it is considered as counting from the end of the
            samples.

        end : int, optional (default=0)
            The ending index for the range of indices to evaluate splits for.
            If negative, it is considered as counting from the end of the
            samples.

        indices : iterable of int, optional (default=None)
            The indices to evaluate splits for. If None, the range of indices
            from `start + 1` to `end` will be used.

        Returns
        -------
        dict[str, Number | np.ndarray]
            A dictionary containing the result of evaluating splits for all
            positions in the specified range of indices. The keys of the
            dictionary are the names of the criterion measures, and the values
            are scalars for node-wise info or arrays for results in each
            position.
        """
        start = start if start >= 0 else self.y.shape[0] + start
        end = end if end > 0 else self.y.shape[0] + end

        if self.sample_weight is None:
            weighted_n_samples = float(self.y.shape[0])
            weighted_n_node_samples = float(end - start)
        else:
            weighted_n_samples = self.sample_weight.sum()
            weighted_n_node_samples = self.sample_weight[start:end].sum()

        if indices is None:
            indices = range(start + 1, end)

        result = defaultdict(list)

        for pos in indices:
            split = self.evaluate_split(
                pos=pos,
                start=start,
                end=end,
                weighted_n_samples=weighted_n_samples,
                weighted_n_node_samples=weighted_n_node_samples,
            )
            for k, v in split.items():
                result[k].append(v)

        result = {k: np.array(v) for k, v in result.items()}
        result['weighted_n_samples'] = weighted_n_samples
        result['weighted_n_node_samples'] = weighted_n_node_samples

        return result


class ReferenceGini(BaseReferenceCriterion):
    def impurity(self, y):
        sq_sums = 0
        for col in y.T:
            _, counts = np.unique(col, return_counts=True)
            sq_sums += np.square(counts / y.shape[0]).sum()
        return 1 - sq_sums / y.shape[1]


class ReferenceEntropy(BaseReferenceCriterion):
    def impurity(self, y):
        col_entropies = 0
        for col in y.T:
            _, counts = np.unique(col, return_counts=True)
            p = counts / y.shape[0]
            col_entropies -= (p * np.log2(p)).sum()
        return col_entropies / y.shape[1]


class ReferenceSquaredError(BaseReferenceCriterion):
    def impurity(self, y):
        return y.var(0).mean()


class ReferenceSquaredErrorGSO(BaseReferenceCriterion):
    def impurity(self, y):
        return y.var()


class ReferenceFriedman(ReferenceSquaredError):
    def impurity_improvement(self, split_data):
        total_sum_left = self.y[:split_data['pos']].sum()
        total_sum_right = self.y[split_data['pos']:].sum()

        diff = (
            split_data['weighted_n_right'] * total_sum_left
            - split_data['weighted_n_left'] * total_sum_right
        )

        return diff ** 2 / (
            split_data['weighted_n_right']
            * split_data['weighted_n_left']
            * split_data['weighted_n_node_samples']
            * self.y.shape[1] ** 2
        )


class ReferenceFriedmanGSO(ReferenceSquaredErrorGSO):
    def impurity_improvement(self, split_data):
        total_sum_left = self.y[split_data['start']:split_data['pos']].sum()
        total_sum_right = self.y[split_data['pos']:split_data['end']].sum()

        diff = (
            split_data['weighted_n_right'] * total_sum_left
            - split_data['weighted_n_left'] * total_sum_right
        )

        return diff ** 2 / (
            split_data['weighted_n_right']
            * split_data['weighted_n_left']
            * split_data['weighted_n_node_samples']
            * self.y.shape[1]  # weighted_n_node_cols
        )


class ReferenceMeanDistance(BaseReferenceCriterion):
    def impurity(self, y):
        if y.shape[0] != y.shape[1]:
            raise ValueError(f'y must be square. Received {y.shape=}')
        if y.shape[0] <= 1:
            return 0.0
        return y[np.triu_indices(y.shape[0], 1)].mean()

    def children_impurity(self, y_region, rel_pos):
        y_left = y_region[:rel_pos, :rel_pos]
        y_right = y_region[rel_pos:, rel_pos:]
        return (
            self.impurity(y_left),
            self.impurity(y_right),
        )
    
    def evaluate_split(self, *args, **kwargs):
        split = super().evaluate_split(*args, **kwargs)

        if self.sample_weight is None:
            split['weighted_n_left'] = (
                (split['weighted_n_left'] - 1)
                * split['weighted_n_left'] / 2
            )
            split['weighted_n_right'] = (
                (split['weighted_n_right'] - 1)
                * split['weighted_n_right'] / 2
            )
            split['weighted_n_samples'] = (
                (split['weighted_n_samples'] - 1)
                * split['weighted_n_samples'] / 2
            )
            split['weighted_n_node_samples'] = (
                (split['weighted_n_node_samples'] - 1)
                * split['weighted_n_node_samples'] / 2
            )
        else:
            raise NotImplementedError
        
        return split


class ReferenceCompositeSS(BaseReferenceCriterion):
    def __init__(
        self,
        sup_criterion,
        unsup_criterion,
        average_both_axes=False,
    ):
        self.sup_criterion = sup_criterion
        self.unsup_criterion = unsup_criterion
        self.average_both_axes = average_both_axes
        self.sup_criterion.average_both_axes = average_both_axes

    def set_data(self, X, y, supervision, sample_weight=None):
        self.X = X
        self.y = y
        self.sample_weight = sample_weight
        self.supervision = supervision
        self.sup_criterion.set_data(X, y, sample_weight)
        self.unsup_criterion.set_data(X, X, sample_weight)

    def evaluate_split(
        self,
        *args,
        **kwargs,
    ) -> Dict[str, int | float]:

        sup_split = self.sup_criterion.evaluate_split(*args, **kwargs)
        unsup_split = self.unsup_criterion.evaluate_split(*args, **kwargs)

        # sup_split['root_impurity'] = sup_split['impurity_parent']
        # unsup_split['root_impurity'] = unsup_split['impurity_parent']

        unsup_split['improvement'] = \
            self.unsup_criterion.impurity_improvement(unsup_split)
        sup_split['improvement'] = \
            self.sup_criterion.impurity_improvement(sup_split)

        result = {
            **{'supervised_' + k: v for k, v in sup_split.items()},
            **{'unsupervised_' + k: v for k, v in unsup_split.items()},
        }

        result['impurity_parent'] = 1.0

        result['impurity_left'] = self.combine_impurities(
            sup_split['impurity_left'] / sup_split['impurity_parent'],
            unsup_split['impurity_left'] / unsup_split['impurity_parent'],
        )
        result['impurity_right'] = self.combine_impurities(
            sup_split['impurity_right'] / sup_split['impurity_parent'],
            unsup_split['impurity_right'] / unsup_split['impurity_parent'],
        )

        result['improvement'] = self.impurity_improvement(result)
        result['pos'] = sup_split['pos']

        return result

    def combine_impurities(self, sup_impurity, unsup_impurity):
        # If the node's Y partition is homogeneous (all values are equal), we
        # disregard the unsupervised criterion to avoid further redundant
        # splitting (all descendant nodes would have the same output value).
        if sup_impurity == 0:
            unsup_impurity = 0

        if self.average_both_axes:
            # unsupervised impurity of the other axis is 1.0, since it does
            # not change.
            #  imp = imp + imp_other_axis
            unsup_impurity += 1
            unsup_impurity /= 2

        return (
            self.supervision * sup_impurity
            + (1 - self.supervision) * unsup_impurity
        )

    def impurity_improvement(self, split_data: dict):
        sup_improvement = (
            self.supervision
            * split_data['supervised_improvement']
            / split_data['supervised_impurity_parent']
        )
        unsup_improvement = (
            + (1 - self.supervision)
            * split_data['unsupervised_improvement']
            / split_data['unsupervised_impurity_parent']
        )
        if self.average_both_axes:  # The improvement on the other axis is zero
            unsup_improvement /= 2

        return sup_improvement + unsup_improvement

    def impurity(self):
        return None


# =============================================================================
# Criterion fixtures and corresponding impurity functions
# =============================================================================


@pytest.fixture(
    params=[
        (
            make_semisupervised_criterion,
            {
                'supervised_criterion': UnsupervisedGini,
                'unsupervised_criterion': UnsupervisedSquaredError,
                'ss_class': SSCompositeCriterion,
            },
            ReferenceCompositeSS(ReferenceGini(), ReferenceSquaredError()),
        ),
        (
            make_semisupervised_criterion,
            {
                'supervised_criterion': UnsupervisedEntropy,
                'unsupervised_criterion': UnsupervisedSquaredError,
                'ss_class': SSCompositeCriterion,
            },
            ReferenceCompositeSS(ReferenceEntropy(), ReferenceSquaredError()),
        ),
        (
            make_semisupervised_criterion,
            {
                'supervised_criterion': UnsupervisedSquaredError,
                'unsupervised_criterion': UnsupervisedSquaredError,
            },
            ReferenceCompositeSS(
                ReferenceSquaredError(),
                ReferenceSquaredError(),
            ),
        ),
        (
            make_semisupervised_criterion,
            {
                'supervised_criterion': UnsupervisedFriedman,
                'unsupervised_criterion': UnsupervisedFriedman,
            },
            ReferenceCompositeSS(
                ReferenceFriedman(),
                ReferenceFriedman(),
            ),
        ),
        (
            make_semisupervised_criterion,
            {
                'supervised_criterion': UnsupervisedFriedman,
                'unsupervised_criterion': UnsupervisedSquaredError,
            },
            ReferenceCompositeSS(
                ReferenceFriedman(),
                ReferenceSquaredError(),
            ),
        ),
        (
            make_semisupervised_criterion,
            {
                'supervised_criterion': UnsupervisedSquaredError,
                'unsupervised_criterion': MeanDistance,
            },
            ReferenceCompositeSS(
                ReferenceSquaredError(),
                ReferenceMeanDistance(),
            ),
        ),
    ],
    ids=[
        'gini|mse',
        'entropy|mse',
        'mse',
        'friedman',
        'friedman|mse',
        'mse|mean_distance',
    ],
)
def semisupervised_criterion(
    request,
    n_samples,
    n_features,
    n_classes,
    supervision,
):
    factory, args, ref_criterion = request.param

    default_criterion_args = dict(
        n_features=n_features[0],
        n_samples=n_samples[0],
        n_classes=n_classes[0],
        n_outputs=n_samples[1],
        supervision=supervision,
    )

    return {
        'criterion': factory(**default_criterion_args | args),
        'ref_criterion': ref_criterion,
    }


@pytest.fixture(
    params=[
        (
            make_2dss_criterion,
            {
                'supervised_criteria': AxisGini,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'ss_criteria': SSCompositeCriterion,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceCompositeSS(
                ReferenceGini(),
                ReferenceSquaredError(),
                average_both_axes=True,
            ),
        ),
        (
            make_2dss_criterion,
            {
                'supervised_criteria': AxisEntropy,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'ss_criteria': SSCompositeCriterion,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceCompositeSS(
                ReferenceEntropy(),
                ReferenceSquaredError(),
                average_both_axes=True,
            ),
        ),
        (
            make_2dss_criterion,
            {
                'supervised_criteria': AxisSquaredErrorGSO,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'ss_criteria': SSCompositeCriterion,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceCompositeSS(
                ReferenceSquaredErrorGSO(),
                ReferenceSquaredError(),
                average_both_axes=True,
            ),
        ),
        (
            make_2dss_criterion,
            {
                'supervised_criteria': AxisSquaredError,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'ss_criteria': SSCompositeCriterion,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceCompositeSS(
                ReferenceSquaredError(),
                ReferenceSquaredError(),
                average_both_axes=True,
            ),
        ),
        (
            make_2dss_criterion,
            {
                'supervised_criteria': AxisFriedmanGSO,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'ss_criteria': SSCompositeCriterion,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceCompositeSS(
                ReferenceFriedmanGSO(),
                ReferenceSquaredError(),
                average_both_axes=True,
            ),
        ),
        (
            make_2dss_criterion,
            {
                'supervised_criteria': AxisSquaredErrorGSO,
                'unsupervised_criteria': MeanDistance,
                'ss_criteria': SSCompositeCriterion,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceCompositeSS(
                ReferenceSquaredErrorGSO(),
                ReferenceMeanDistance(),
                average_both_axes=True,
            ),
        ),
    ],
    ids=[
        'gini|mse',
        'entropy|mse',
        'mse_gso',
        'mse_gmo',
        'friedman|mse',
        'mse_gso|mean_distance',
    ],
)
def ss_bipartite_criterion(
    request,
    n_samples,
    n_features,
    supervision,
    n_classes,
):
    factory, args, ref_criterion = request.param

    default_criterion_args = dict(
        n_features=n_features,
        n_samples=n_samples,
        n_classes=n_classes,
        n_outputs=n_samples[::-1],
        supervision=supervision,
    )

    return {
        'criterion': factory(**default_criterion_args | args),
        'ref_criterion': ref_criterion,
    }


@pytest.fixture(
    params=[
        (
            make_bipartite_criterion,
            {
                'criteria': AxisGini,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceGini(average_both_axes=True),
        ),
        (
            make_bipartite_criterion,
            {
                'criteria': AxisEntropy,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceEntropy(average_both_axes=True),
        ),
        (
            make_bipartite_criterion,
            {
                'criteria': AxisSquaredErrorGSO,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceSquaredErrorGSO(),
        ),
        (
            make_bipartite_criterion,
            {
                'criteria': AxisSquaredError,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceSquaredError(average_both_axes=True),
        ),
        (
            make_bipartite_criterion,
            {
                'criteria': AxisFriedmanGSO,
                'bipartite_criterion_class': GMOSA,
            },
            ReferenceFriedmanGSO(average_both_axes=True),
        ),
    ],
    ids=['gini', 'entropy', 'mse_gso', 'mse_gmo', 'friedman'],
)
def supervised_bipartite_criterion(request, n_samples, n_classes):
    factory, args, ref_criterion = request.param

    default_criterion_args = dict(
        n_samples=n_samples,
        n_outputs=n_samples[::-1],
        n_classes=n_classes,
    )
    return {
        'criterion': factory(**default_criterion_args | args),
        'ref_criterion': ref_criterion,
    }


@pytest.fixture
def semisupervised_splits(
    semisupervised_criterion,
    data,
    supervision,
):
    return get_ss_splits(
        ss_criterion_data=semisupervised_criterion,
        data=data,
        supervision=supervision,
        apply_criterion=apply_ss_criterion,
    )


@pytest.fixture
def supervised_splits(
    criterion,
    data,
    n_samples,
):
    return get_ss_splits(
        ss_criterion_data=criterion,
        data=data,
        supervision=None,
        apply_criterion=apply_criterion,
    )


@pytest.fixture
def ss_bipartite_splits(
    ss_bipartite_criterion,
    data,
    n_samples,
    supervision,
):
    return get_ss_bipartite_splits(
        ss_bipartite_criterion=ss_bipartite_criterion,
        data=data,
        n_samples=n_samples,
        supervision=supervision,
        apply_criterion=apply_bipartite_ss_criterion,
    )


@pytest.fixture
def supervised_bipartite_splits(
    supervised_bipartite_criterion,
    n_samples,
    data,
):
    return get_ss_bipartite_splits(
        ss_bipartite_criterion=supervised_bipartite_criterion,
        data=data,
        n_samples=n_samples,
        supervision=None,
        apply_criterion=apply_bipartite_criterion,
    )


# =============================================================================
# Utility functions
# =============================================================================


def turn_into_classification(y):
    n_bins = 5
    y = np.digitize(y, np.linspace(y.min(), y.max(), n_bins))
    return y.astype('float64')


def turn_into_pairwise(x):
    return rbf_kernel(x)


def get_ss_splits(
    *,
    ss_criterion_data,
    data,
    supervision,
    apply_criterion,
):
    XX, Y, *_ = data
    X = XX[0].astype('float64')

    criterion = ss_criterion_data['criterion']
    ref_criterion = ss_criterion_data['ref_criterion']

    if (
        type(criterion) in CLASSIFICATION_CRITERIA
        or supervision is not None
        and type(criterion.supervised_criterion) in CLASSIFICATION_CRITERIA
    ):
        Y = turn_into_classification(Y)

    if (
        supervision is not None
        and type(criterion.unsupervised_criterion) in PAIRWISE_CRITERIA
    ):
        X = turn_into_pairwise(X)

    criterion.set_X(X)
    split_data = apply_criterion(criterion, Y)

    ref_criterion.set_data(X, Y, supervision=supervision)
    ref_split_data = ref_criterion.eval_all_splits()

    return {'splits': split_data, 'reference_splits': ref_split_data}


def get_ss_bipartite_splits(
    *,
    ss_bipartite_criterion,
    data,
    n_samples,
    supervision,
    apply_criterion,
):
    X, Y, *_ = data

    start_row, start_col = 0, 0
    end_row, end_col = n_samples

    criterion = ss_bipartite_criterion['criterion']
    ref_criterion = ss_bipartite_criterion['ref_criterion']

    if (
        type(criterion.criterion_rows) in CLASSIFICATION_CRITERIA
        or supervision is not None
        and type(criterion.criterion_rows.supervised_criterion)
        in CLASSIFICATION_CRITERIA
    ):
        Y = turn_into_classification(Y)

    if supervision is not None:
        if (
            type(criterion.criterion_rows.unsupervised_criterion)
            in PAIRWISE_CRITERIA
        ):
            X[0] = turn_into_pairwise(X[0])
        if (
            type(criterion.criterion_cols.unsupervised_criterion)
            in PAIRWISE_CRITERIA
        ):
            X[1] = turn_into_pairwise(X[1])

    row_splits, col_splits = apply_criterion(
        criterion,
        X[0],
        X[1],
        Y,
        start=[start_row, start_col],
        end=[end_row, end_col],
    )

    ref_criterion.set_data(X[0], Y, supervision=supervision)
    ref_row_splits = ref_criterion.eval_all_splits(
        start=start_row, end=end_row,
    )

    ref_criterion.set_data(X[1], Y.T, supervision=supervision)
    ref_col_splits = ref_criterion.eval_all_splits(
        start=start_col, end=end_col,
    )

    return {
        'splits': (row_splits, col_splits),
        'reference_splits': (ref_row_splits, ref_col_splits),
    }


def assert_correct_proxy_factors(ss_splits, **kwargs):
    sup_data = scipy.stats.linregress(
        ss_splits['supervised_improvement'],
        ss_splits['supervised_proxy_improvement'],
    )
    unsup_data = scipy.stats.linregress(
        ss_splits['unsupervised_improvement'],
        ss_splits['unsupervised_proxy_improvement'],
    )

    assert sup_data.pvalue < 1e-3, "Uncorrelated supervised proxies."
    assert unsup_data.pvalue < 1e-3, "Uncorrelated unsupervised proxies."
    assert np.unique(ss_splits['supervised_proxy_factor']).shape[0] == 1, (
        "Supervised proxy factor is not constant!"
    )
    assert np.unique(ss_splits['unsupervised_proxy_factor']).shape[0] == 1, (
        "Unsupervised proxy factor is not constant!"
    )

    assert_equal_dicts(
        {
            'sup_proxy_factor': ss_splits['supervised_proxy_factor'][0],
            'unsup_proxy_factor': ss_splits['unsupervised_proxy_factor'][0],
        },
        {
            'sup_proxy_factor': sup_data.slope,
            'unsup_proxy_factor': unsup_data.slope,
        },
        **kwargs,
    )


def assert_ranks_are_close(a, b, **kwargs):
    a_s = np.argsort(a)
    b_s = np.argsort(b)
    # If any of swapped values (between both b and a ordering) is not
    # different, do not raise.
    try:
        assert_equal_dicts(
            {'a_sorted': a[a_s], 'b_sorted': b[a_s]},
            {'a_sorted': a[b_s], 'b_sorted': b[b_s]},
            **kwargs,
        )
    except AssertionError as error:
        if "'b_sorted' and 'a_sorted" in str(error):
            raise


def sort_by_feature(x, y, feature):
    sorted_indices = x[:, feature].argsort()
    return x[sorted_indices], y[sorted_indices]


# =============================================================================
# Tests
# =============================================================================


def test_no_axis_criterion_as_unsupervised():
    with pytest.raises(TypeError, match=r'.*AxisCriterion'):
        make_semisupervised_criterion(
            supervised_criterion=AxisSquaredError,
            unsupervised_criterion=AxisSquaredError,
            n_features=100,
            n_samples=200,
            n_outputs=10,
            supervision=0.5,
        )


def test_ss_criterion(semisupervised_splits):
    rtol = 1e-7
    atol = 1e-7

    ref_splits = semisupervised_splits['reference_splits']
    ref_splits['ss_improvement'] = ref_splits['improvement']
    ref_splits['ss_impurity_parent'] = ref_splits['impurity_parent']
    ref_splits['ss_impurity_right'] = ref_splits['impurity_right']
    ref_splits['ss_impurity_left'] = ref_splits['impurity_left']

    assert_equal_dicts(
        semisupervised_splits['splits'],
        ref_splits,
        rtol=rtol,
        atol=atol,
        ignore={
            'n_samples',
            'n_node_samples',
            'proxy_improvement',
            'n_outputs',
            'end',
            'start',
        },
        subset=semisupervised_splits.keys(),
        differing_keys='raise',
    )


def test_ss_criterion_proxies(semisupervised_splits):
    """Test semi-supervised proxy_impurities.

    Test that the order of the proxy improvement values matches the order of
    the final improvement values.
    """
    rtol = 1e-7
    atol = 1e-7

    splits = semisupervised_splits['splits']
    ref_splits = semisupervised_splits['reference_splits']

    assert_ranks_are_close(
        splits['supervised_proxy_improvement'],
        splits['supervised_improvement'],
        msg_prefix='(supervised proxy) ',
        rtol=rtol,
        atol=atol,
    )
    assert_ranks_are_close(
        splits['unsupervised_proxy_improvement'],
        splits['unsupervised_improvement'],
        msg_prefix='(unsupervised proxy) ',
        rtol=rtol,
        atol=atol,
    )
    assert_ranks_are_close(
        splits['ss_proxy_improvement'],
        splits['ss_improvement'],
        msg_prefix='(proxy vs. improvement) ',
        rtol=rtol,
        atol=atol,
    )
    assert_ranks_are_close(
        splits['ss_proxy_improvement'],
        ref_splits['improvement'],
        msg_prefix='(proxy vs. reference) ',
        rtol=rtol,
        atol=atol,
    )


def test_supervised_criterion(data):
    XX, Y, *_ = data
    X = XX[0]

    criterion = UnsupervisedSquaredError(
        n_samples=Y.shape[0],
        n_outputs=Y.shape[1],
    )
    split_data = apply_criterion(criterion, Y)

    ref_criterion = ReferenceSquaredError()
    ref_criterion.set_data(X, Y)
    ref_split_data = ref_criterion.eval_all_splits()

    assert_equal_dicts(
        split_data,
        ref_split_data,
        rtol=1e-7,
        atol=1e-8,
        differing_keys='raise',
        ignore={
            'proxy_improvement',
            'n_outputs',
            'n_samples',
            'n_node_samples',
            'start',
            'end',
        },
    )

    # Test that the order of the proxy improvement values matches the order of
    # the final improvement values.
    assert_ranks_are_close(
        split_data['proxy_improvement'],
        ref_split_data['improvement'],
        msg_prefix='(proxy) ',
    )


def test_ss_bipartite_criterion(ss_bipartite_splits):
    rtol = 1e-7
    atol = 1e-7

    row_splits, col_splits = ss_bipartite_splits['splits']
    ref_row_splits, ref_col_splits = ss_bipartite_splits['reference_splits']

    ref_row_splits['axis_weighted_n_samples'] = \
        ref_row_splits['weighted_n_samples']
    ref_col_splits['axis_weighted_n_samples'] = \
        ref_col_splits['weighted_n_samples']
    ref_row_splits['axis_weighted_n_node_samples'] = \
        ref_row_splits['weighted_n_node_samples']
    ref_col_splits['axis_weighted_n_node_samples'] = \
        ref_col_splits['weighted_n_node_samples']

    ignore = {
        'n_samples',
        'n_node_samples',
        'supervised_n_samples',
        'supervised_n_node_samples',
        'unsupervised_n_samples',
        'unsupervised_n_node_samples',
        'weighted_n_samples',
        'weighted_n_node_samples',
    }

    assert_equal_dicts(
        row_splits,
        ref_row_splits,
        msg_prefix='(rows) ',
        subset=ref_row_splits.keys(),
        ignore=ignore,
        rtol=rtol,
        atol=atol,
        differing_keys='raise',
    )
    assert_equal_dicts(
        col_splits,
        ref_col_splits,
        msg_prefix='(cols) ',
        subset=ref_row_splits.keys(),
        ignore=ignore,
        rtol=rtol,
        atol=atol,
        differing_keys='raise',
    )


def test_ss_bipartite_proxy_improvement(ss_bipartite_splits):
    """Test bipartite semi-supervised proxy_impurities.

    Test that the order of the proxy improvement values matches the order of
    the final improvement values.
    """
    # Test that the order of the proxy improvement values matches the order of
    # the final improvement values.
    rtol = 1e-7
    atol = 1e-7

    for axis, (splits, ref_splits) in enumerate(zip(
        ss_bipartite_splits['splits'],
        ss_bipartite_splits['reference_splits'],
    )):
        # Perform control test asserting that the impurity improvement matches
        # the manually calculated reference values.
        assert_ranks_are_close(
            splits['improvement'],
            ref_splits['improvement'],
            msg_prefix=f'({axis=}, bipartite vs reference) ',
            rtol=rtol,
            atol=atol,
        )

        # Additional tests for better inspection of semisupervised splits.
        if 'ss_improvement' in splits:
            assert_ranks_are_close(
                splits['ss_improvement'],
                splits['improvement'],
                msg_prefix=f'({axis=}, axis vs bipartite) ',
                rtol=rtol,
                atol=atol,
            )
            assert_ranks_are_close(
                splits['supervised_proxy_improvement'],
                splits['supervised_improvement'],
                msg_prefix=f'({axis=}, sup proxy vs axis) ',
                rtol=rtol,
                atol=atol,
            )
            assert_ranks_are_close(
                splits['unsupervised_proxy_improvement'],
                splits['unsupervised_improvement'],
                msg_prefix=f'({axis=}, unsup proxy vs axis) ',
                rtol=rtol,
                atol=atol,
            )
            assert_ranks_are_close(
                splits['ss_proxy_improvement'],
                splits['ss_improvement'],
                msg_prefix=f'({axis=}, proxy vs axis) ',
                rtol=rtol,
                atol=atol,
            )

        # Final main assertion.
        assert_ranks_are_close(
            splits['proxy_improvement'],
            ref_splits['improvement'],
            msg_prefix=f'({axis=}, proxy vs reference) ',
            rtol=rtol,
            atol=atol,
        )


def test_supervised_bipartite_criterion(supervised_bipartite_splits):
    test_ss_bipartite_criterion(supervised_bipartite_splits)


def test_supervised_bipartite_proxy_improvement(supervised_bipartite_splits):
    test_ss_bipartite_proxy_improvement(supervised_bipartite_splits)


def test_monopartite_proxy_factors(semisupervised_splits):
    assert_correct_proxy_factors(semisupervised_splits['splits'])


def test_bipartite_proxy_factors(ss_bipartite_splits):
    row_splits, col_splits = ss_bipartite_splits['splits']
    assert_correct_proxy_factors(row_splits, msg_prefix='(rows) ')
    assert_correct_proxy_factors(col_splits, msg_prefix='(cols) ')


@pytest.mark.parametrize(
    'single_output_ref_criterion, multioutput_ref_criterion', [
        (ReferenceSquaredErrorGSO(), ReferenceSquaredError()),
        (ReferenceFriedmanGSO(), ReferenceFriedman()),
        (
            ReferenceCompositeSS(
                ReferenceSquaredErrorGSO(),
                ReferenceSquaredError(),
            ),
            ReferenceCompositeSS(
                ReferenceSquaredError(),
                ReferenceSquaredError(),
            ),
        )
    ],
    ids=['mse', 'friedman', 'ss_mse'],
)
def test_gso_gmo_equivalence(
    data,
    n_samples,
    supervision,
    single_output_ref_criterion,
    multioutput_ref_criterion,
):
    """Tests if a single-output impurity metric is invariant in the GSO format.

    The test ensures that a single-output version of a metric (that ignores
    different columns as different outputs) yields the same values on the two
    formats of a bipartite dataset described bellow.

        1. Global Multi-Output (GMO) format: 
            Axis 0 (rows) receives:
                X = X[0]
                Y = Y
            Axis 1 (columns) receives:
                X = X[1]
                Y = Y.T

        2. Global Single Output (GSO) format: 
            Axis 0 (rows) receives:
                X = np.repeat(X[0], n_cols, axis=0)
                Y = Y.reshape(-1, 1))
            Axis 1 (columns) receives:
                X = np.repeat(X[1], n_rows, axis=0)
                Y = Y.reshape(-1, 1))

    An example of single-output impurity is the UnsupervisedSquaredError defined as Y.var() instead
    of its more usual multioutput form Y.var(0).mean().

    If the metric passes this test, we can employ it in testing bipartite
    criteria directly in the more convenient GMO format (done by
    `test_semisupervised_bipartite_criterion()`) being sure the criterion being compared to
    the metric would yield the same results if the multi-output counterpart of
    the metric were to be applied in the GSO-formatted dataset.
    """
    start_row, start_col = 0, 0
    end_row, end_col = n_samples
    n_rows, n_cols = n_samples

    rtol = 1e-7
    atol = 1e-7

    X, Y, *_ = data

    # Split positions to evaluate
    row_indices = np.arange(start_row + 1, end_row)
    col_indices = np.arange(start_col + 1, end_col)

    multioutput_ref_criterion.set_data(
        X=np.repeat(X[0], n_cols, axis=0),
        y=Y.reshape(-1, 1),
        supervision=supervision,
    )
    mo_row_splits = multioutput_ref_criterion.eval_all_splits(
        indices=row_indices * n_cols,
        start=start_row * n_cols,
        end=end_row * n_cols,
    )

    multioutput_ref_criterion.set_data(
        X=np.repeat(X[1], n_rows, axis=0),
        y=Y.T.reshape(-1, 1),
        supervision=supervision,
    )
    mo_col_splits = multioutput_ref_criterion.eval_all_splits(
        indices=col_indices * n_rows,
        start=start_col * n_rows,
        end=end_col * n_rows,
    )

    single_output_ref_criterion.set_data(X[0], Y, supervision=supervision)
    so_row_splits = single_output_ref_criterion.eval_all_splits(
        start=start_row,
        end=end_row,
    )

    single_output_ref_criterion.set_data(X[1], Y.T, supervision=supervision)
    so_col_splits = single_output_ref_criterion.eval_all_splits(
        start=start_col,
        end=end_col,
    )

    mo_row_splits['pos'] = row_indices
    mo_col_splits['pos'] = col_indices

    ignore = {
        'n_samples',
        'n_node_samples',
        'weighted_n_left',
        'weighted_n_right',
        'weighted_n_samples',
        'weighted_n_node_samples',
        'supervised_n_samples',
        'supervised_n_node_samples',
        'supervised_weighted_n_left',
        'supervised_weighted_n_right',
        'supervised_weighted_n_samples',
        'supervised_weighted_n_node_samples',
        'unsupervised_n_samples',
        'unsupervised_n_node_samples',
        'unsupervised_weighted_n_left',
        'unsupervised_weighted_n_right',
        'unsupervised_weighted_n_samples',
        'unsupervised_weighted_n_node_samples',
        'supervised_pos',
        'unsupervised_pos',
        'start',
        'end',
        'supervised_start',
        'supervised_end',
        'unsupervised_start',
        'unsupervised_end',
    }
    assert_equal_dicts(
        so_row_splits,
        mo_row_splits,
        msg_prefix='(rows) ',
        ignore=ignore,
        rtol=rtol,
        atol=atol,
        differing_keys='raise',
    )
    assert_equal_dicts(
        so_col_splits,
        mo_col_splits,
        msg_prefix='(cols) ',
        ignore=ignore,
        rtol=rtol,
        atol=atol,
        differing_keys='raise',
    )



import itertools
import copy
from numbers import Integral
import numpy as np
from sklearn.utils._param_validation import validate_params, Interval
from sklearn.model_selection._split import (
    BaseCrossValidator,
    KFold,
    StratifiedKFold,
    ShuffleSplit,
    StratifiedShuffleSplit,
    _validate_shuffle_split,
)
from sklearn.utils.validation import _num_samples

from bipartite_learn.datasets.base import BipartiteDataset


class MultipartiteCrossValidator(BaseCrossValidator):
    @validate_params(
        {
            "cross_validators": [BaseCrossValidator, list, tuple],
            "diagonal": ["boolean"],
            "n_parts": [Interval(Integral, 2, None, closed="left")],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, cross_validators, diagonal=False, n_parts=2):
        if isinstance(cross_validators, BaseCrossValidator):
            self.cross_validators = [
                copy.deepcopy(cross_validators) for _ in range(n_parts)
            ]
            self.n_parts = n_parts
        else:
            self.n_parts = len(cross_validators)
            self.cross_validators = cross_validators

        self.diagonal = diagonal

    def get_n_splits(self, X=None, y=None, groups=None):
        return self._check_n_splits(X, y, groups)

    def _check_n_splits(self, X=None, y=None, groups=None):
        ax_n_splits = list(
            self._get_from_children_cv("get_n_splits", X, y, groups),
        )

        if self.diagonal:
            # Ensure all cross validators report the same number of splits.
            g = itertools.groupby(ax_n_splits)
            next(g, None)

            # if not all elements in ax_n_splits are equal:
            if next(g, False):
                raise ValueError(
                    "Cross-validators must generate the same number"
                    " of splits if diagonal=True"
                )
            return ax_n_splits[0]

        return np.prod(ax_n_splits)

    def split(self, X, y=None, groups=None):
        train_test = self._get_from_children_cv("split", X, y=y, groups=groups)
        # Each train_test element represents a fold, containing train/test
        # indices for all axes:
        #
        # next(train_test) == (
        #    (train_fold0_ax0, test_fold0_ax0),
        #    (train_fold0_ax1, test_fold0_ax1),
        #    (train_fold0_ax2, test_fold0_ax2),
        #    ...
        # ),
        # train_test.shape = n_folds, n_axis, 2

        # So train_test is
        #   (all_train/test_tuples for fold1,
        #    all_train/test_tuples for fold2,

        combine_func = zip if self.diagonal else itertools.product
        split_iter = combine_func(*train_test)
        # split_iter.shape = n_axis, 2, n_folds

        # For each axis, pick one train/test tuple from a diferent fold.
        # next(splits_iter) == (
        #   ((train_ax0_fold_a, test_ax0_fold_a),
        #    (train_ax1_fold_c, test_ax1_fold_c),
        #    (train_ax2_fold_h, test_ax2_fold_h),
        #    (train_ax3_fold_b, test_ax3_fold_b),
        #    ...
        #   ),

        # Put in another way:
        # next(splits_iter) == (train_test_ax1, train_test_ax2, ...)

        for split in split_iter:
            yield split

    def _iter_test_masks(self, X=None, y=None, groups=None):
        iter_ = self._get_from_children_cv(
            "_iter_test_masks",
            X,
            y=y,
            groups=groups,
        )

        for item in iter_:
            yield item

    def _iter_test_indices(self, X=None, y=None, groups=None):
        iter_ = self._get_from_children_cv(
            "_iter_test_indices",
            X,
            y=y,
            groups=groups,
        )

        for item in iter_:
            yield item

    def _get_from_children_cv(self, attr, X=None, y=None, groups=None):
        if X is None:
            X = [None] * self.n_parts
        if groups is None:
            groups = [None] * self.n_parts
        if y is None:
            y = [None] * self.n_parts
        else:
            # FIXME: is np.moveaxis well suited here?
            y = [np.moveaxis(y, ax, 0) for ax in range(self.n_parts)]

        for cv, Xi, yi, group in zip(self.cross_validators, X, y, groups):
            yield getattr(cv, attr)(Xi, yi, group)


class HoldOut(BaseCrossValidator):
    def __init__(self, test_size=None, train_size=None):
        self.test_size = test_size
        self.train_size = train_size
        self._default_test_size = 0.25

    def get_n_splits(self, X=None, y=None, groups=None):
        return 1

    def _iter_test_indices(self, X, y=None, groups=None):
        n_samples = _num_samples(X)
        n_train, n_test = _validate_shuffle_split(
            n_samples,
            self.test_size,
            self.train_size,
            default_test_size=self._default_test_size,
        )
        yield np.arange(n_train, n_samples)


def _check_train_test_combinations(
    ttc,
    n_parts,
    include_train=None,
    symbols="LT",
    sep="",
):
    """Check train_test_combinations parameter."""
    # ttc = train_test_combinations
    include_train = include_train or True
    names_provided = (ttc is not None) and (
        isinstance(ttc[0], str) or isinstance(ttc[0][0], str)
    )

    if ttc is None:
        ret_ttc = itertools.product((0, 1), repeat=n_parts)
        if not include_train:
            next(ret_ttc)  # Skip (0, 0, ...), only training data.
        ret_ttc = list(ret_ttc)

    elif any(len(c) != 2 for c in ttc):
        raise ValueError(
            "train_test_combinations must contain only 2-lengthed sequences."
        )

    # If ttc contains string or list-like of strings, translate.
    elif names_provided:
        names = ttc
        ret_ttc = [[symbols.index(s) for s in train_test] for train_test in names]

    if not names_provided:
        names = [[symbols[is_test] for is_test in train_test] for train_test in ret_ttc]

    names = [n[0] + sep + n[1] for n in names]
    return ret_ttc, names


def _repeat_if_single(n_parts, *args):
    result = []
    for arg in args:
        if not isinstance(arg, (tuple, list)):  # TODO: Better criteria?
            arg = [copy.deepcopy(arg) for _ in range(n_parts)]
        result.append(arg)
    return result


def combine_train_test_indices(
    train_test,  # iterable of (train_indices, test_indices) tuples for each axis.
    train_test_combinations=None,
    include_train=None,
):
    n_parts = len(train_test)

    train_test_combinations, train_test_names = _check_train_test_combinations(
        ttc=train_test_combinations,
        n_parts=n_parts,
        include_train=include_train,
    )

    train_test_index_combinations = {}

    # NOTE: ttc stands for train-test combinations
    for is_test_tuple, ttc_name in zip(train_test_combinations, train_test_names):
        # is_test_tuple ~= (0, 1, 1, 0)
        test_indices = [
            ax_train_test[is_test]
            for is_test, ax_train_test in zip(is_test_tuple, train_test)
        ]
        train_test_index_combinations[ttc_name] = test_indices

    return train_test_index_combinations


def make_multipartite_kfold(
    cv=None,
    shuffle=False,
    stratified=False,
    random_state=None,
    diagonal=False,
    n_parts=None,
):
    """Build a n_parts-partite KFold cross-validator.

    Wraps n_parts KFold or StratifiedKFold cross-validators with a
    MultipartiteCrossValidator, that provides train-test indices for splitting
    InputDataND objects. The returned wrapper object is compatible with
    parameter search objects in bipartite_learn.model_selection, such as
    MultipartiteGridSearchCV and MultipartiteRandomizedSearchCV.

    The parameters 'cv', 'shuffle' and 'stratified' can be
    optionally passed as a tuple or list of values intended for each respective
    axis of the ND data. Otherwise, the single value is equally considered for
    all axes.

    Parameters
    ----------
    cv : int or CrossValidator or list-like of these, default=None
        TODO
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    random_state : int, RandomState instance or None, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.
    shuffle : bool or list-like of bool, default=True
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    stratified : bool or list-like of array-like, default=None
        whether or not to generate a stratified splitter.
        Read more in the :ref:`User Guide <stratification>`.
        One can optionally pass arrays for each axis in a list or tuple. If a
        single array is provided, it will be considered for all axes.
    n_parts : int
        If none of the other parameters is a list or a sequence, you must indi-
        cate the number of dimensions.

    Returns
    -------
    cross-validation splitter : MultipartiteCrossValidator
    """
    if n_parts is None:
        if isinstance(cv, (list, tuple)):
            n_parts = len(cv)
        elif isinstance(shuffle, (list, tuple)):
            n_parts = len(shuffle)
        elif isinstance(stratified, (list, tuple)):
            n_parts = len(stratified)
        else:
            raise ValueError(
                "If none of cv, shuffle or stratified is tuple"
                " or list, n_parts must be provided."
            )

    (
        cv,
        shuffle,
        stratified,
    ) = _repeat_if_single(n_parts, cv, shuffle, stratified)

    cv_list = []

    for ax in range(n_parts):
        if not isinstance(cv[ax], BaseCrossValidator):
            CVClass = StratifiedKFold if stratified[ax] else KFold
            cv[ax] = CVClass(
                n_splits=cv[ax],
                random_state=random_state,
                shuffle=shuffle[ax],
            )
        cv_list.append(copy.deepcopy(cv[ax]))

    return check_multipartite_cv(cv_list, n_parts=n_parts, diagonal=diagonal)


def make_multipartite_train_test_splitter(
    n_parts,
    test_size=None,
    train_size=None,
    shuffle=False,
    stratified=False,
    random_state=None,
):
    """Build a len(n_samples)-dimensional train-test splitter.

    Wraps len(n_samples) CrossValidator data splitters with a
    MultipartiteCrossValidator, that provides train-test indices for splitting
    InputDataND objects. The returned wrapper object is compatible with
    parameter search objects in bipartite_learn.model_selection, such as
    MultipartiteGridSearchCV and MultipartiteRandomizedSearchCV.

    The parameters 'test_size', 'train_size', 'shuffle' and 'stratify' can be
    optionally passed as a tuple or list of values intended for each respective
    axis of the ND data. Otherwise, the single value is equally considered for
    all axes.

    Parameters
    ----------
    test_size : float or int or list-like of (float or int), default=None
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. If ``train_size`` is also None, it will
        be set to 0.25.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    train_size : float or int or list-like of (float or int), default=None
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    random_state : int, RandomState instance or None, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.
    shuffle : bool or list-like of bool, default=True
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    stratified : bool or list-like of array-like, default=None
        whether or not to generate a stratified splitter.
        Read more in the :ref:`User Guide <stratification>`.
        One can optionally pass arrays for each axis in a list or tuple. If a
        single array is provided, it will be considered for all axes.

    Returns
    -------
    cross-validation splitter : MultipartiteCrossValidator
    """
    (
        test_size,
        train_size,
        shuffle,
        stratified,
    ) = _repeat_if_single(n_parts, test_size, train_size, shuffle, stratified)

    cv_list = []

    for ax in range(n_parts):
        if shuffle[ax]:
            if stratified[ax]:
                CVClass = StratifiedShuffleSplit
            else:
                CVClass = ShuffleSplit

            cv = CVClass(
                test_size=test_size[ax],
                train_size=train_size[ax],
                random_state=random_state,
                n_splits=1,
            )
        else:
            if stratified[ax]:
                raise ValueError(
                    "Stratified train/test split is not implemented for"
                    "shuffle=False"
                )
            CVClass = HoldOut
            cv = CVClass(
                test_size=test_size[ax],
                train_size=train_size[ax],
            )
        cv_list.append(copy.deepcopy(cv))

    return check_multipartite_cv(cv_list)


def check_multipartite_cv(
    cv,
    y=None,
    *,
    n_parts=2,  # Overwritten by cv if cv is list/tuple
    classifier=False,
    diagonal=False,
    shuffle=False,
    random_state=None,
):
    """
    Convert cross-validation objects into a `MultipartiteCrossValidator`.

    Parameters
    ----------
    cv : int, cross-validator, or tuple/list with such objects
        Determines the cross-validation splitting strategy. Possible inputs for
        cv are:

        - integer: is converted to a `KFold` instance with the provided number
          of folds.
        - A scikit-learn cross-validator instance.
        - A list or tuple of integers or cross-validation objects, specifying
          one of such objects for each axis of the multipartite dataset.

    y : array-like, default=None
        The interaction matrix (or tensor) that gathers the target values of
        a multipartite dataset.

    n_parts : int, default=2
        The number of parts in the multipartite dataset. E.g. 2 for a bipartite
        task.

    classifier : boolean, default=False
        Whether the task is a classification task. This parameter is deprecated
        and is not used in the function.

    diagonal : boolean, default=False
        Whether to perform diagonal cross-validation. See the :ref:`User Guide`
        for details.

    shuffle : boolean, default=False
        The `shuffle` parameter to be passed to KFold objects created from
        integers.

    random_state : int or RandomState instance, default=None
        The `random_state` parameter to be passed to KFold objects created from
        integers.

    Returns
    -------
    cv : MultipartiteCrossValidator
        A `MultipartiteCrossValidator` to split multipartite data.

    Notes
    -----
    The `shuffle` and `random_state` parameters are only used when `cv` is an
    integer or contains integers.
    """
    if isinstance(cv, MultipartiteCrossValidator):
        return cv

    # TODO: StratifiedKFold is not implemented for multipartite data.
    if isinstance(cv, int):
        cv = KFold(n_splits=cv, shuffle=shuffle, random_state=random_state)

    if isinstance(cv, (tuple, list)):
        cv = [
            KFold(n_splits=i, shuffle=shuffle, random_state=random_state)
            if isinstance(i, int)
            else i
            for i in cv
        ]

    return MultipartiteCrossValidator(
        cv,
        diagonal=diagonal,
        n_parts=n_parts,  # Overritten by cv if cv is list/tuple
    )


def multipartite_train_test_split(
    *data,
    test_size=None,
    train_size=None,
    random_state=None,
    shuffle=True,
    stratify=None,
    input_type=BipartiteDataset,
):
    """Split datasets into random train and test subsets.

    Quick utility that wraps input validation and
    ``next(ShuffleSplit().split(X, y))`` and application to input data
    into a single call for splitting (and optionally subsampling) data in a
    oneliner.
    Read more in the :ref:`User Guide <cross_validation>`.

    The parameters 'test_size', 'train_size', 'shuffle' and 'stratify' can be
    optionally passed as a tuple or list of values intended for each respective
    axis of the ND data. Otherwise, the single value is equally considered for
    all axes.

    Parameters
    ----------
    *data : list-like of (InputDataND or (X, y) list-like pairs)
        N-dimensional datasets to be splitted.
    test_size : float or int or list-like of (float or int), default=None
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. If ``train_size`` is also None, it will
        be set to 0.25.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    train_size : float or int or list-like of (float or int), default=None
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    random_state : int, RandomState instance or None, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.
    shuffle : bool or list-like of bool, default=True
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.
        One can optionally pass values for each axis in a list or tuple. If a
        single value is provided, it will be considered for all axes.
    stratify : array-like or list-like of array-like, default=None
        If not None, data is split in a stratified fashion, using this as
        the class labels.
        Read more in the :ref:`User Guide <stratification>`.
        One can optionally pass arrays for each axis in a list or tuple. If a
        single array is provided, it will be considered for all axes.
    input_type : type, default=BipartiteDataset
        The input data format, with n_samples and n_features attributes.
        Read more in the :ref:`User Guide <input_types>` (TODO).

    Returns
    -------
    splitting : dict,
        Each value will be a list of the sampled datasets provided.
        Each key indentifies if train or test indices were used on each axis.
        E.g. for a 2D dataset, "LT" key means learned (training) rows and
        test columns. "TL" means test rows and training columns. "TT" means the
        data were selected with test rows and test columns as well.
    """
    if stratify is not None:
        raise NotImplementedError("Stratification not implemented.")
    n_sets = len(data)
    if n_sets == 0:
        raise ValueError("At least one item required as input")

    data = list(data)  # We use assignment.

    for i in range(len(data)):
        if not isinstance(data[i], input_type):
            data[i] = input_type(*data[i])

    n_parts = data[0].n_parts
    n_samples = data[0].n_samples

    # n_train, n_test = _validate_shuffle_split(
    #     n_samples=n_samples[ax],
    #     test_size=test_size[ax],
    #     train_size=train_size[ax],
    #     default_test_size=0.25
    # )

    cv = make_multipartite_train_test_splitter(
        n_parts=n_parts,
        test_size=test_size,
        train_size=train_size,
        shuffle=shuffle,
        stratified=stratify is not None,
        random_state=random_state,
    )
    train_test = next(cv.split(X=data[0].X, y=stratify))
    ttc_indices = combine_train_test_indices(train_test, include_train=True)

    split_data = {}
    for ttc_name, ttc_index in ttc_indices.items():
        if n_sets == 1:
            sample = data[0][ttc_index].Xy
        else:
            sample = [d[ttc_index].Xy for d in data]
        split_data[ttc_name] = sample

    return split_data



import logging
import numpy as np
import pytest
from scipy.stats import ttest_ind, ttest_rel
from sklearn.utils._testing import assert_allclose
from sklearn.utils.validation import check_random_state
from sklearn.dummy import DummyRegressor
from sklearn.ensemble._gb import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
)
from sklearn.ensemble import (
    RandomForestRegressor,
    ExtraTreesRegressor,
)
from bipartite_learn.ensemble._gb import (
    BipartiteGradientBoostingClassifier,
    BipartiteGradientBoostingRegressor,
)
from bipartite_learn.ensemble._forest import (
    BipartiteRandomForestRegressor,
    BipartiteExtraTreesRegressor,
)
from bipartite_learn.wrappers import GlobalSingleOutputWrapper
from .utils.make_examples import make_interaction_regression
from .utils.test_utils import assert_equal_dicts, stopwatch
from sklearn import metrics


@pytest.mark.parametrize(
    "monopartite_estimator,bipartite_estimator,common_estimator_params",
    [
        (
            RandomForestRegressor,
            BipartiteRandomForestRegressor,
            dict(
                criterion='squared_error',
                n_estimators=10,
                bootstrap=False,  # Bootstrapping in bipartite data is different.
            ),
        ),
        (
            ExtraTreesRegressor,
            BipartiteExtraTreesRegressor,
            dict(
                criterion='squared_error',
                n_estimators=30,
            ),
        ),
        (
            GradientBoostingRegressor,
            BipartiteGradientBoostingRegressor,
            dict(
                criterion='friedman_mse',
                subsample=0.2,
                n_estimators=10,
            ),
        ),
    ],
)
def test_compare_estimators(
    monopartite_estimator,
    bipartite_estimator,
    common_estimator_params,
    scorer=metrics.mean_squared_error,
    random_state=0,
    n_seeds=30,
):
    logging.info(f"{random_state=}")
    seeds = check_random_state(random_state).randint(
        np.iinfo(np.uint32).max,
        size=n_seeds,
        dtype=np.uint32,
    )
    monopartite_scores = []
    bipartite_scores = []
    dummy_scores = []

    for seed in seeds:
        logging.info(f"{seed=}")
        common_estimator_params.update({'random_state': seed})
        dummy_estimator = DummyRegressor()
        monopartite_estimator_ = monopartite_estimator(**common_estimator_params)
        bipartite_estimator_ = bipartite_estimator(**common_estimator_params)

        XX, Y, X, y = make_interaction_regression(
            n_samples=(40, 50),
            n_features=(9, 10),
            return_molten=True,
            random_state=seed,
            # max_target=2,
        )
        # Classification
        # Y = Y.astype(int)
        # y = y.astype(int)
        bipartite_estimator_.fit(XX, Y)
        monopartite_estimator_.fit(X, y)
        dummy_estimator.fit(X, y)
        
        assert_equal_dicts(
            monopartite_estimator_.get_params(),
            bipartite_estimator_.get_params(),
        )

        dummy_pred = dummy_estimator.predict(X)
        pred1 = monopartite_estimator_.predict(X)
        pred2 = bipartite_estimator_.predict(XX)
        # score1 = np.mean((y - pred1)**2)
        # score2 = np.mean((y - pred2)**2)
        # score1 = monopartite_estimator_.score(X, y)
        # score2 = bipartite_estimator_.score(XX, y)
        dummy_score = scorer(dummy_pred, y)
        score1 = scorer(pred1, y)
        score2 = scorer(pred2, y)

        logging.info(f"Monopartite score: {score1}")
        logging.info(f"Bipartite score:   {score2}")

        dummy_scores.append(dummy_score)
        monopartite_scores.append(score1)
        bipartite_scores.append(score2)
    
    # FIXME: paired test seems more adequate, but does not yield significant results.
    assert ttest_ind(dummy_scores, monopartite_scores).pvalue < 0.05
    assert ttest_ind(dummy_scores, bipartite_scores).pvalue < 0.05
    assert ttest_ind(monopartite_scores, bipartite_scores).pvalue > 0.05



"""
The :mod:`sklearn.ensemble` module includes ensemble-based methods for
classification, regression and anomaly detection.
"""
from ._forest import (
    BipartiteRandomForestRegressor,
    BipartiteExtraTreesRegressor,
)
from ._semisupervised_forest import (
    RandomForestRegressorSS,
    ExtraTreesRegressorSS,
    BipartiteRandomForestRegressorSS,
    BipartiteExtraTreesRegressorSS,
)
from ._gb import (
    BipartiteGradientBoostingRegressor,
    BipartiteGradientBoostingClassifier,
)

__all__ = [
    "BipartiteRandomForestRegressor",
    "BipartiteExtraTreesRegressor",
    "BipartiteGradientBoostingRegressor",
    "BipartiteGradientBoostingClassifier",
    "RandomForestRegressorSS",
    "ExtraTreesRegressorSS",
    "BipartiteRandomForestRegressorSS",
    "BipartiteExtraTreesRegressorSS",
]



from __future__ import annotations
from typing import Callable, Literal
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.neighbors import kneighbors_graph
from sklearn.utils.validation import check_symmetric
from imblearn.base import BaseSampler
from ..utils import check_similarity_matrix

_GammaScaleOptions = Literal["constant", "squares", "squared_errors", "size"]


def nearest_positive_semidefinite(X):
    """Get nearest (Frobenius norm) positive semidefinite matrix from A.

    See Equations (2.1) and (2.2) of [1].
    Also see [https://stackoverflow.com/q/43238173/11286509].

    [1] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    """
    A = (X + X.T)/2
    eigval, eigvec = np.linalg.eig(A)
    eigval[eigval < 0] = 0
    return (eigvec * eigval) @ eigvec.T


def enforce_positive_semidefiniteness(X, tol=1e-5):
    """Enforce positive-semidefiniteness of kernel matrix.

    Modifies only the main diagonal of X.

    Adds to the main diagonal the absolute value of the minimum negative
    eigen-value, returning a positive-semidefinite version of X.

    Parameters
    ----------
    X : square two-dimensional ndarray
        Kernel matrix to transform
    tol : float, default=1e-5
        To avoid small negative values due to numerical precision,
        tol is also added to the diagonal

    Returns
    -------
    ndarray with same shape as X
        Symmetric positive-semidefinite version of X.
    """
    X = (X + X.T) / 2
    assert (X >= 0).all(), "Kernel matrix must be >= 0"
    eigvals = np.linalg.eigvals(X)
    min_negative_eigval = eigvals[eigvals < 0].min()
    np.fill_diagonal(X, X.diagonal() - min_negative_eigval + tol)
    return X


def _scale_gamma(X, gamma: float, gamma_scale: str) -> float:
    """Multiply gamma by an X-dependent factor.
    Parameters
    ----------
    X : 2d ndarray
        Input array of kernel function.
    gamma : float
        `gamma` parameter to be passed to kernel function.
    gamma_scale : {'constant', 'squares', 'squared_errors', 'size'}
        If not 'constant', divide `gamma` by `S / y.shape[0]`, where
        `S = (y**2).sum()`, if `gamma_scale='squares'`,
        `((y-y.mean()) ** 2).sum()` if 'squared_errors' and 'y.size' if 'size'.
    Returns
    -------
    scaled gamma : float
    """
    if gamma_scale == "constant":
        return gamma
    elif gamma_scale == "squares":
        return gamma * X.shape[0] / (X**2).sum()
    elif gamma_scale == "squared_errors":
        return gamma / (X.shape[1] * X.var())
    elif gamma_scale == "size":
        return gamma / X.shape[1]
    else:
        raise ValueError(
            "Unrecognized gamma_scale: {gamma_scale!r}. Valid options are "
            "'constant', 'squares', 'squared_errors' and 'size'."
        )


class TargetKernelLinearCombiner(BaseSampler):
    """Combines provided similarity matrix X with kernel calculated over y

    X is assumed to be a precomputed kernel matrix. The target kernel will be
    calculated with `sklearn.metrics.pairwise.pairwise_kernels` and combined
    with X simply by taking `alpha*X + (1-alpha)*y_kernel`.

    The default kernel is RBF, so that it calculates the 'gaussian interaction
    profile' as described by [1].

    Valid values for metric are:
        ['additive_chi2', 'chi2', 'linear', 'poly', 'polynomial', 'rbf',
        'laplacian', 'sigmoid', 'cosine']

    Parameters
    ----------
    alpha : float, default=0.5
        Controls the fraction of the target information in the linear
        combination with the provided similarities. alpha=1 means no change,
        alpha=0 means no original X data will remain.
    metric : str or callable, default="rbf"
        The metric to use when calculating kernel between instances in a
        feature array. If metric is a string, it must be one of the metrics
        in `sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS`.
        If metric is "precomputed", y is assumed to be a kernel matrix.
        Alternatively, if metric is a callable function, it is called on each
        pair of instances (rows) and the resulting value recorded. The callable
        should take two rows from y as input and return the corresponding
        kernel value as a single number. This means that callables from
        :mod:`sklearn.metrics.pairwise` are not allowed, as they operate on
        matrices, not single samples. Use the string identifying the kernel
        instead.
    gamma : float, default=1.0
        `gamma` parameter of kernel function if metric is callable, 'chi2',
        'polynomial', 'rbf', 'laplacian' or 'sigmoid'.
    gamma_scale : {'constant', 'squares', 'squared_errors', 'size'}, \
    default='squares'
        If not 'constant', divide `gamma` by `S / y.shape[0]`, where
        `S = (y**2).sum()`, if `gamma_scale='squares'`,
        `((y-y.mean()) ** 2).sum()` if 'squared_errors' and 'y.size' if 'size'.
    filter_params : bool, default=False
        Whether to filter invalid kernel parameters or not.
    n_jobs : int, default=None
        The number of jobs to use for the kernel computation. This works by
        breaking down the y matrix into n_jobs even slices and computing
        them in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
    **kwds : optional keyword parameters
        Any further parameters are passed directly to the kernel function.
    References
    ----------
    .. [1] :doi:`"Gaussian interaction profile kernels for predicting drug–\
       target interaction" <https://doi.org/10.1093/bioinformatics/btr500>`
       van Laarhoven, Nabuurs and Marchiori, 2011.
    """
    def __init__(
        self,
        alpha: float = 0.5,
        metric: str | Callable = "rbf",
        gamma: float = 1.0,
        gamma_scale: _GammaScaleOptions = "squares",
        filter_params: bool = False,
        n_jobs: int | None = None,
        **kernel_params,
    ):
        self.alpha = alpha
        self.metric = metric
        self.gamma = gamma
        self.gamma_scale = gamma_scale
        self.filter_params = filter_params
        self.n_jobs = n_jobs
        self._kernel_params = kernel_params

    def _fit_resample(self, X, y, **fit_params):
        self.effective_gamma_ = _scale_gamma(y, self.gamma, self.gamma_scale)
        y_kernel = pairwise_kernels(
            y,
            metric=self.metric,
            gamma=self.effective_gamma_,
            filter_params=self.filter_params,
            n_jobs=self.n_jobs,
            **self._kernel_params,
        )
        return self.alpha * X + (1-self.alpha) * y_kernel, y


class TargetKernelDiffuser(BaseSampler):
    """Calculates kernel on y and performs non-linear kernel diffusion.

    DOI: https://doi.org/10.1016/j.aca.2016.01.014
    Hao _et al._, 2016.

    X is assumed to be a precomputed kernel matrix. The target kernel will be
    calculated with `sklearn.metrics.pairwise.pairwise_kernels` and combined
    with X by a kernel diffusion procedure [1].

    The default kernel is RBF, so that it calculates the 'gaussian interaction
    profile' as described by [2].

    Valid values for metric are:
        ['additive_chi2', 'chi2', 'linear', 'poly', 'polynomial', 'rbf',
        'laplacian', 'sigmoid', 'cosine']

    Parameters
    ----------
    n_iter : int, default=2
        Number of diffusion iterations.
    n_neighbors : int, default=4
        n_neighbors parameter passed to kneighbors_graph for local similarity
        calculation.
    metric : str or callable, default="rbf"
        The metric to use when calculating kernel between instances in a
        feature array. If metric is a string, it must be one of the metrics
        in `sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS`.
        If metric is "precomputed", y is assumed to be a kernel matrix.
        Alternatively, if metric is a callable function, it is called on each
        pair of instances (rows) and the resulting value recorded. The callable
        should take two rows from y as input and return the corresponding
        kernel value as a single number. This means that callables from
        :mod:`sklearn.metrics.pairwise` are not allowed, as they operate on
        matrices, not single samples. Use the string identifying the kernel
        instead.
    gamma : float, default=1.0
        `gamma` parameter of kernel function if metric is callable, 'chi2',
        'polynomial', 'rbf', 'laplacian' or 'sigmoid'.
    gamma_scale : {'constant', 'squares', 'squared_errors', 'size'}, \
    default='squares'
        If not 'constant', divide `gamma` by `S / y.shape[0]`, where
        `S = (y**2).sum()`, if `gamma_scale='squares'`,
        `((y-y.mean()) ** 2).sum()` if 'squared_errors' and 'y.size' if 'size'.
    filter_params : bool, default=False
        Whether to filter invalid kernel parameters or not.
    n_jobs : int, default=None
        The number of jobs to use for the kernel computation. This works by
        breaking down the y matrix into n_jobs even slices and computing
        them in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
    **kwds : optional keyword parameters
        Any further parameters are passed directly to the kernel function.
    References
    ----------
    .. [1] :doi:`"Improved prediction of drug-target interactions using \
       regularized least squares integrating with kernel fusion technique" \
        <https://doi.org/10.1016/j.aca.2016.01.014>`
        Ming Hao, Yanli Wang and Bryant, Stephen H, 2016.
    .. [2] :doi:`"Gaussian interaction profile kernels for predicting drug–\
       target interaction" <https://doi.org/10.1093/bioinformatics/btr500>`
       van Laarhoven, Nabuurs and Marchiori, 2011.
    """
    def __init__(
        self,
        n_iter: int = 2,
        n_neighbors: int = 4,
        metric: str | Callable = "rbf",
        gamma: float = 1.0,
        gamma_scale: _GammaScaleOptions = "squares",
        filter_params: bool = False,
        n_jobs: int | None = None,
        **kwds,
    ):
        self.n_iter = n_iter
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.gamma = gamma
        self.gamma_scale = gamma_scale
        self.filter_params = filter_params
        self.n_jobs = n_jobs
        self._kernel_params = kwds

    def _fit_resample(self, X, y):
        y_kernel = pairwise_kernels(
            y,
            metric=self.metric,
            gamma=_scale_gamma(y, self.gamma, self.gamma_scale),
            filter_params=self.filter_params,
            n_jobs=self.n_jobs,
            **self._kernel_params,
        )
        S = X.copy()  # Similarity matrix
        S_net = y_kernel.copy()  # Connections-based similarity

        self._normalize_rows(S)
        self._normalize_rows(S_net)
        self._symmetrize(S)
        self._symmetrize(S_net)

        S_local = self._local_graph(S, self.n_neighbors)
        S_net_local = self._local_graph(S_net, self.n_neighbors)

        for _ in range(self.n_iter):
            breakpoint()
            S = S_local @ S_net @ S_local.T
            S_net = S_net_local @ S @ S_net_local.T
            self._symmetrize(S)
            self._symmetrize(S_net)

        S_final = (S + S_net) / 2

        self._normalize_rows(S_final)
        self._symmetrize(S_final)

        # NOTE: Not in the original paper.
        S_final /= S_final.max(axis=1, keepdims=True)  # Normalize to [0, 1]

        return S_final, y

    @staticmethod
    def _symmetrize(S):
        """Symmetrize and add 1 to diagonal to encourage positive-definiteness.
        """
        S += S.T
        S /= 2
        np.fill_diagonal(S, S.diagonal() + 1)

    @staticmethod
    def _normalize_rows(S):
        """Divide each value by its row's sum, so that each row sums to 1."""
        S /= S.sum(axis=1, keepdims=True)

    @staticmethod
    def _local_graph(S, n_neighbors):
        S_local = kneighbors_graph(
            2 - S,  # Convert similarity to distance
            n_neighbors=n_neighbors,
            metric="precomputed",
            mode="distance",
        )  # sparse matrix
        # Convert non-zero entries back to similarities
        S_local.data = 2 - S_local.data
        # Normalize rows
        S_local /= S_local.sum(axis=1)  # Dimension is kept by default
        return np.array(S_local)


class SymmetryEnforcer(BaseSampler):
    """Make matrix symmetric by averaging it with its transpose.
    """
    def _fit_resample(self, X, y):
        return check_symmetric(X), y

class PositiveSemidefiniteEnforcer(BaseEstimator, TransformerMixin):
    """Modify main diagonal to enforce positive semidefiniteness.

    Adds to the main diagonal the absolute value of the minimum negative
    eigen-value, returning a positive-semidefinite version of X.

    Parameters
    ----------
    tol : float, default=1e-5
        To avoid small negative values due to numerical precision,
        tol is also added to the diagonal.
    """
    def __init__(self, tol=1e-5):
        self.tol = tol

    def fit(self, X, y=None, **fit_params):
        return self

    def fit_transform(self, X, y=None, **fit_params):
        return enforce_positive_semidefiniteness(X, self.tol)


class SimilarityDistanceSwitcher(BaseEstimator, TransformerMixin):
    """Transforms x into (1 - x)."""
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return 1 - check_similarity_matrix(X, check_symmetry=False)

    def inverse_transform(self, X):
        return self.transform(X)



# cython: boundscheck=False
import copy
import warnings
from sklearn.tree._splitter cimport Splitter
from sklearn.tree._criterion cimport RegressionCriterion, Criterion

import numpy as np
cimport numpy as np

np.import_array()
cdef double INFINITY = np.inf


cdef inline void _init_split(
    MultipartiteSplitRecord* self,
    SIZE_t start_pos,
    SIZE_t axis
) nogil:
    self.axis = axis
    self.split_record.impurity_left = INFINITY
    self.split_record.impurity_right = INFINITY
    self.split_record.pos = start_pos
    self.split_record.feature = 0
    self.split_record.threshold = 0.
    self.split_record.improvement = -INFINITY


cdef class BipartiteSplitter:
    """PBCT splitter implementation.

    Wrapper class to coordinate one Splitter for each axis in a two-dimensional
    problem as described by Pliakos _et al._ (2018).
    """
    def __cinit__(
        self,
        Splitter splitter_rows,
        Splitter splitter_cols,
        BipartiteCriterion bipartite_criterion,
        min_samples_leaf,
        min_weight_leaf,
    ):
        """Store each axis' splitter."""
        self.n_samples = 0
        self.weighted_n_samples = 0.0
        self.splitter_rows = splitter_rows
        self.splitter_cols = splitter_cols
        self.bipartite_criterion = bipartite_criterion

        self.min_samples_leaf = min_samples_leaf
        self.min_weight_leaf = min_weight_leaf
        self.min_rows_leaf = self.splitter_rows.min_samples_leaf
        self.min_cols_leaf = self.splitter_cols.min_samples_leaf

    def __reduce__(self):
        return (
            type(self),
            (
                self.splitter_rows,
                self.splitter_cols,
                self.bipartite_criterion,
                self.min_samples_leaf,
                self.min_weight_leaf,
            ),
            self.__getstate__(),
        )

    def __getstate__(self):
        return {}

    def __setstate__(self, d):
        pass

    cdef int init(
        self,
        object X,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,  # TODO: test sample_weight
    ) except -1:
        """Initialize the axes' splitters.
        Take in the input data X, the target Y, and optional sample weights.
        Use them to initialize each axis' splitter.
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.

        Parameters
        ----------
        X : object
            This contains the inputs. It is a list-like of 2d numpy arrays, one
            array for each axis.
        y : ndarray, dtype=DOUBLE_t
            This is the matrix of targets, or true labels, for the samples
        sample_weight : (DOUBLE_t*)[2]
            The weights of the row and column samples, where higher weighted
            samples are fit closer than lower weight samples. If not provided,
            all samples are assumed to have uniform weight.
        """
        self.y = y
        self.n_rows = y.shape[0]
        self.n_cols = y.shape[1]

        self.X_rows = X[0]
        self.X_cols = X[1]

        self.n_row_features = self.X_rows.shape[1]

        if sample_weight is None:
            self.row_weights = None
            self.col_weights = None
        else:
            # First self.n_rows sample weights refer to rows, the others
            # refer to columns.
            self.row_weights = sample_weight
            self.col_weights = sample_weight[self.n_rows:]

        # splitters do not receive y, the Bipartite criterion is who takes
        # charge of calling Criterion.init() passing y to the criterion.
        # TODO: missing values is not supported yet (4th parameter=None).
        self.splitter_rows.init(X[0], None, self.row_weights, None)
        self.splitter_cols.init(X[1], None, self.col_weights, None)
        self.row_indices = self.splitter_rows.samples
        self.col_indices = self.splitter_cols.samples

        self.n_samples = self.n_rows * self.n_cols
        self.weighted_n_rows = self.splitter_rows.weighted_n_samples
        self.weighted_n_cols = self.splitter_cols.weighted_n_samples
        self.weighted_n_samples = self.weighted_n_rows * self.weighted_n_cols

        return 0

    cdef int node_reset(
        self,
        SIZE_t[2] start,
        SIZE_t[2] end,
        double* weighted_n_node_samples
    ) nogil except -1:
        """Reset splitters on node samples[start[0]:end[0], start[1]:end[1]].
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.

        Parameters
        ----------
        start : SIZE_t[2]
            The (i, j) indices of the first sample to consider
        end : SIZE_t[2]
            The (i, j) indices of the last sample to consider
        weighted_n_node_samples[2] : ndarray, dtype=double pointer
            The total weight of those samples
        """
        cdef SIZE_t n_node_rows, n_node_cols
        cdef SIZE_t eff_min_rows_leaf
        cdef SIZE_t eff_min_cols_leaf

        n_node_rows = end[0] - start[0]
        n_node_cols = end[1] - start[1]

        # Ceil division.
        eff_min_rows_leaf = 1 + (self.min_samples_leaf-1) // n_node_cols
        eff_min_cols_leaf = 1 + (self.min_samples_leaf-1) // n_node_rows

        self.splitter_rows.min_samples_leaf = max(
            self.min_rows_leaf, eff_min_rows_leaf)
        self.splitter_cols.min_samples_leaf = max(
            self.min_cols_leaf, eff_min_cols_leaf)

        self.splitter_rows.start = start[0]
        self.splitter_rows.end = end[0]
        self.splitter_cols.start = start[1]
        self.splitter_cols.end = end[1]

        self.start[0], self.start[1] = start[0], start[1]
        self.end[0], self.end[1] = end[0], end[1]

        self.bipartite_criterion.init(
            self.X_rows,
            self.X_cols,
            self.y,
            self.row_weights,
            self.col_weights,
            self.weighted_n_rows,
            self.weighted_n_cols,
            self.row_indices,
            self.col_indices,
            start,
            end,
        )

        weighted_n_node_samples[0] = (
            self.bipartite_criterion.weighted_n_node_samples
        )
        weighted_n_node_samples[1] = (
            self.bipartite_criterion.weighted_n_node_rows
        )
        weighted_n_node_samples[2] = (
            self.bipartite_criterion.weighted_n_node_cols
        )

        return 0

    cdef int node_split(
        self,
        double impurity,
        MultipartiteSplitRecord* split,
        SIZE_t[2] n_constant_features
    ) nogil except -1:
        """Find the best split on node samples.
        It should return -1 upon errors.
        """
        cdef SIZE_t ax  # axis, 0 == rows, 1 == columns
        cdef MultipartiteSplitRecord current_split, best_split
        cdef DOUBLE_t imp_left, imp_right

        # Using Cython casting trick to iterate over splitters without the GIL
        cdef (void*)[2] splitters
        splitters[0] = <void*> self.splitter_rows
        splitters[1] = <void*> self.splitter_cols

        _init_split(&best_split, self.splitter_rows.end, 0)

        # TODO: No need to "reorganize into samples" in each axis.
        # (sklearn.tree._splitter.pyx, line 417)
        for ax in range(2):
            if (
                (self.end[ax] - self.start[ax])
                < 2 * (<Splitter>splitters[ax]).min_samples_leaf
            ):
                continue  # min_samples_leaf not satisfied.

            (<Splitter>splitters[ax]).node_split(
                impurity,
                <SplitRecord*> &current_split,
                &n_constant_features[ax],
            )

            # When no valid split has been found, the child splitter sets
            # the split position at the end of the current node.
            if current_split.split_record.pos == self.end[ax]:
                continue

            self.bipartite_criterion.children_impurity(
                &imp_left, &imp_right, axis=ax,
            )
            imp_improve = self.bipartite_criterion.impurity_improvement(
                impurity, imp_left, imp_right, axis=ax,
            )

            if imp_improve > best_split.split_record.improvement:
                best_split = current_split
                best_split.split_record.improvement = imp_improve
                best_split.split_record.impurity_left = imp_left
                best_split.split_record.impurity_right = imp_right

                best_split.axis = ax
                if ax == 1:
                    best_split.split_record.feature += self.n_row_features

        split[0] = best_split

    cdef void node_value(self, double* dest) nogil:
        """Copy the value (prototype) of node samples into dest."""
        self.bipartite_criterion.node_value(dest)

    cdef double node_impurity(self) nogil:
        """Return the impurity of the current node."""
        return self.bipartite_criterion.node_impurity()



"""
The :mod:`bipartite_learn.model_selection._validation` module includes classes
and functions to validate the model.
"""
# Author: Pedro Ilídio <ilidio@alumni.usp.br>
# Modified from scikit-learn.
# License: BSD 3 clause

import warnings
import numbers
import time
from traceback import format_exc

import numpy as np
from joblib import logger

import sklearn.utils
from sklearn.model_selection._validation import (
    _score,
    _normalize_score_results,
    _warn_or_raise_about_fit_failures,
    _aggregate_score_dicts,
)
from sklearn.utils.validation import (
    _make_indexable,
)
from sklearn.utils._tags import _safe_tags
from sklearn.base import is_classifier, clone
from sklearn.utils.parallel import delayed, Parallel
from sklearn.metrics import check_scoring
from sklearn.metrics._scorer import _check_multimetric_scoring

from ._split import check_multipartite_cv, _check_train_test_combinations


__all__ = [
    "multipartite_cross_validate",
]


def multipartite_cross_validate(
    estimator,
    X,
    y=None,
    *,
    groups=None,
    scoring=None,
    cv=None,
    n_jobs=None,
    verbose=0,
    fit_params=None,
    pre_dispatch="2*n_jobs",
    return_train_score=False,
    return_estimator=False,
    error_score=np.nan,
    # bipartite_learn-exclusive params:
    diagonal=False,
    train_test_combinations=None,
    pairwise=False,  # possibility to override estimator._more_tags['pairwise']
    shuffle=False,
    random_state=None,
):
    # TODO: ND adapt docs.
    """Evaluate metric(s) by cross-validation and also record fit/score times.

    Read more in the :ref:`User Guide <multimetric_cross_validation>`.

    Parameters
    ----------
    estimator : estimator object implementing 'fit'
        The object to use to fit the data.

    X : array-like of shape (n_samples, n_features)
        The data to fit. Can be for example a list, or an array.

    y : array-like of shape (n_samples,) or (n_samples, n_outputs), \
            default=None
        The target variable to try to predict in the case of
        supervised learning.

    groups : array-like of shape (n_samples,), default=None
        Group labels for the samples used while splitting the dataset into
        train/test set. Only used in conjunction with a "Group" :term:`cv`
        instance (e.g., :class:`GroupKFold`).

    diagonal : bool, default=False
        Wether to combine each axis splits in a product-manner or zip-manner.

    scoring : str, callable, list, tuple, or dict, default=None
        Strategy to evaluate the performance of the cross-validated model on
        the test set.

        If `scoring` represents a single score, one can use:

        - a single string (see :ref:`scoring_parameter`);
        - a callable (see :ref:`scoring`) that returns a single value.

        If `scoring` represents multiple scores, one can use:

        - a list or tuple of unique strings;
        - a callable returning a dictionary where the keys are the metric
          names and the values are the metric scores;
        - a dictionary with metric names as keys and callables a values.

        See :ref:`multimetric_grid_search` for an example.

    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross validation,
        - int, to specify the number of folds in a `(Stratified)KFold`,
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For int/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`.Fold` is used. These splitters are instantiated
        with `shuffle=False` so the splits will be the same across calls.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        .. versionchanged:: 0.22
            ``cv`` default value if None changed from 3-fold to 5-fold.

    n_jobs : int, default=None
        Number of jobs to run in parallel. Training the estimator and computing
        the score are parallelized over the cross-validation splits.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    verbose : int, default=0
        The verbosity level.

    fit_params : dict, default=None
        Parameters to pass to the fit method of the estimator.

    pre_dispatch : int or str, default='2*n_jobs'
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A str, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    return_train_score : bool, default=False
        Whether to include train scores.
        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.

        .. versionadded:: 0.19

        .. versionchanged:: 0.21
            Default value was changed from ``True`` to ``False``

    return_estimator : bool, default=False
        Whether to return the estimators fitted on each split.

        .. versionadded:: 0.20

    error_score : 'raise' or numeric, default=np.nan
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised.
        If a numeric value is given, FitFailedWarning is raised.

        .. versionadded:: 0.20

    Returns
    -------
    scores : dict of float arrays of shape (n_splits,)
        Array of scores of the estimator for each run of the cross validation.

        A dict of arrays containing the score/time arrays for each scorer is
        returned. The possible keys for this ``dict`` are:

            ``test_score``
                The score array for test scores on each cv split.
                Suffix ``_score`` in ``test_score`` changes to a specific
                metric like ``test_r2`` or ``test_auc`` if there are
                multiple scoring metrics in the scoring parameter.
            ``train_score``
                The score array for train scores on each cv split.
                Suffix ``_score`` in ``train_score`` changes to a specific
                metric like ``train_r2`` or ``train_auc`` if there are
                multiple scoring metrics in the scoring parameter.
                This is available only if ``return_train_score`` parameter
                is ``True``.
            ``fit_time``
                The time for fitting the estimator on the train
                set for each cv split.
            ``score_time``
                The time for scoring the estimator on the test set for each
                cv split. (Note time for scoring on the train set is not
                included even if ``return_train_score`` is set to ``True``
            ``estimator``
                The estimator objects for each cv split.
                This is available only if ``return_estimator`` parameter
                is set to ``True``.

    Examples
    --------
    >>> from sklearn import datasets, linear_model
    >>> from sklearn.model_selection import cross_validate
    >>> from sklearn.metrics import make_scorer
    >>> from sklearn.metrics import confusion_matrix
    >>> from sklearn.svm import LinearSVC
    >>> diabetes = datasets.load_diabetes()
    >>> X = diabetes.data[:150]
    >>> y = diabetes.target[:150]
    >>> lasso = linear_model.Lasso()

    Single metric evaluation using ``cross_validate``

    >>> cv_results = cross_validate(lasso, X, y, cv=3)
    >>> sorted(cv_results.keys())
    ['fit_time', 'score_time', 'test_score']
    >>> cv_results['test_score']
    array([0.3315057 , 0.08022103, 0.03531816])

    Multiple metric evaluation using ``cross_validate``
    (please refer the ``scoring`` parameter doc for more information)

    >>> scores = cross_validate(lasso, X, y, cv=3,
    ...                         scoring=('r2', 'neg_mean_squared_error'),
    ...                         return_train_score=True)
    >>> print(scores['test_neg_mean_squared_error'])
    [-3635.5... -3573.3... -6114.7...]
    >>> print(scores['train_r2'])
    [0.28009951 0.3908844  0.22784907]

    See Also
    --------
    cross_val_score : Run cross-validation for single metric evaluation.

    cross_val_predict : Get predictions from each split of cross-validation for
        diagnostic purposes.

    sklearn.metrics.make_scorer : Make a scorer from a performance metric or
        loss function.

    """

    # *X, y, groups = indexable(*X, y, groups)
    # Call _make_indexable on each one instead of indexable. The only difference
    # being we avoid calling sklear.utils.check_consistent_length this way.
    X = [_make_indexable(a) for a in X]
    y = _make_indexable(y)

    ndim = y.ndim

    if groups is None:
        groups = [None] * ndim
    groups = [_make_indexable(a) for a in groups]

    # Check dimension consistency
    if not (ndim == len(X) == len(groups)):
        # FIXME: multi-output: y would have an extra dimension.
        raise ValueError(
            "Incompatible dimensions. One must ensure "
            "y.ndim == len(X) == len(groups)"
        )

    cv = check_multipartite_cv(
        cv,
        y,
        classifier=is_classifier(estimator),
        diagonal=diagonal,
        shuffle=shuffle,
        random_state=random_state,
    )

    (
        train_test_combinations,
        train_test_names,
    ) = _check_train_test_combinations(
        ttc=train_test_combinations,
        n_parts=len(X),
        include_train=return_train_score,
    )

    if callable(scoring):
        scorers = scoring
    elif scoring is None or isinstance(scoring, str):
        scorers = check_scoring(estimator, scoring)
    else:
        scorers = _check_multimetric_scoring(estimator, scoring)

    splits_iter = cv.split(X, y, groups)
    # next(splits_iter) == train_test ==
    #   ((train_ax0, test_ax0), (train_ax1, test_ax1), ...)

    # We clone the estimator to make sure that all the folds are
    # independent, and that it is pickle-able.
    parallel = Parallel(n_jobs=n_jobs, verbose=verbose, pre_dispatch=pre_dispatch)
    results = parallel(
        delayed(_bipartite_fit_and_score)(
            clone(estimator),
            X,
            y,
            scorers,
            train_test,
            verbose,
            None,
            fit_params,
            return_train_score=return_train_score,
            return_times=True,
            return_estimator=return_estimator,
            error_score=error_score,
            train_test_combinations=train_test_combinations,
            train_test_names=train_test_names,
            pairwise=pairwise,
        )
        for train_test in splits_iter
    )

    _warn_or_raise_about_fit_failures(results, error_score)

    # For callabe scoring, the return type is only know after calling. If the
    # return type is a dictionary, the error scores can now be inserted with
    # the correct key.
    #
    # if callable(scoring):  # TODO
    #     _insert_error_scores(results, error_score)

    results = _aggregate_score_dicts(results)

    ret = {}
    ret["fit_time"] = results["fit_time"]
    ret["score_time"] = results["score_time"]

    if return_estimator:
        ret["estimator"] = results["estimator"]

    for ttc_name in train_test_names:
        test_scores_dict = _normalize_score_results(results[f"{ttc_name}_scores"])

        for name in test_scores_dict:
            ret[f"{ttc_name}_{name}"] = test_scores_dict[name]

    return ret


def _bipartite_fit_and_score(
    estimator,
    X,
    y,
    scorer,
    train_test,
    verbose,
    parameters,
    fit_params,
    return_train_score=False,
    return_parameters=False,
    return_n_test_samples=False,
    return_times=False,
    return_estimator=False,
    split_progress=None,
    candidate_progress=None,
    error_score=np.nan,
    # bipartite_learn-specific parameters:
    train_test_combinations=None,
    train_test_names=None,
    pairwise=False,  # Overrides estimator._more_tags['pairwise']
):
    """Fit estimator and compute scores for a given dataset split.

    Parameters
    ----------
    estimator : estimator object implementing 'fit'
        The object to use to fit the data.

    X : array-like of shape (n_samples, n_features)
        The data to fit.

    y : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        The target variable to try to predict in the case of
        supervised learning.

    scorer : A single callable or dict mapping scorer name to the callable
        If it is a single callable, the return value for ``train_scores`` and
        ``test_scores`` is a single float.

        For a dict, it should be one mapping the scorer name to the scorer
        callable object / function.

        The callable object / fn should have signature
        ``scorer(estimator, X, y)``.

    train : array-like of shape (n_train_samples,)
        Indices of training samples.

    test : array-like of shape (n_test_samples,)
        Indices of test samples.

    verbose : int
        The verbosity level.

    error_score : 'raise' or numeric, default=np.nan
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised.
        If a numeric value is given, FitFailedWarning is raised.

    parameters : dict or None
        Parameters to be set on the estimator.

    fit_params : dict or None
        Parameters that will be passed to ``estimator.fit``.

    return_train_score : bool, default=False
        Compute and return score on training set.

    return_parameters : bool, default=False
        Return parameters that has been used for the estimator.

    split_progress : {list, tuple} of int, default=None
        A list or tuple of format (<current_split_id>, <total_num_of_splits>).

    candidate_progress : {list, tuple} of int, default=None
        A list or tuple of format
        (<current_candidate_id>, <total_number_of_candidates>).

    return_n_test_samples : bool, default=False
        Whether to return the ``n_test_samples``.

    return_times : bool, default=False
        Whether to return the fit/score times.

    return_estimator : bool, default=False
        Whether to return the fitted estimator.

    Returns
    -------
    result : dict with the following attributes
        train_scores : dict of scorer name -> float
            Score on training set (for all the scorers),
            returned only if `return_train_score` is `True`.
        test_scores : dict of scorer name -> float
            Score on testing set (for all the scorers).
        n_test_samples : int
            Number of test samples.
        fit_time : float
            Time spent for fitting in seconds.
        score_time : float
            Time spent for scoring in seconds.
        parameters : dict or None
            The parameters that have been evaluated.
        estimator : estimator object
            The fitted estimator.
        fit_error : str or None
            Traceback str if the fit failed, None if the fit succeeded.
    """
    if not isinstance(error_score, numbers.Number) and error_score != "raise":
        raise ValueError(
            "error_score must be the string 'raise' or a numeric value. "
            "(Hint: if using 'raise', please make sure that it has been "
            "spelled correctly.)"
        )

    progress_msg = ""
    if verbose > 2:
        if split_progress is not None:
            progress_msg = f" {split_progress[0]+1}/{split_progress[1]}"
        if candidate_progress and verbose > 9:
            progress_msg += f"; {candidate_progress[0]+1}/{candidate_progress[1]}"

    if verbose > 1:
        if parameters is None:
            params_msg = ""
        else:
            sorted_keys = sorted(parameters)  # Ensure deterministic o/p
            params_msg = ", ".join(f"{k}={parameters[k]}" for k in sorted_keys)
    if verbose > 9:
        start_msg = f"[CV{progress_msg}] START {params_msg}"
        print(f"{start_msg}{(80 - len(start_msg)) * '.'}")

    # Adjust length of sample weights
    fit_params = fit_params if fit_params is not None else {}
    # fit_params = _check_fit_params(X, fit_params, train)  # FIXME

    if parameters is not None:
        # clone after setting parameters in case any parameters
        # are estimators (like pipeline steps)
        # because pipeline doesn't clone steps in fit
        cloned_parameters = {}
        for k, v in parameters.items():
            cloned_parameters[k] = clone(v, safe=False)

        estimator = estimator.set_params(**cloned_parameters)

    start_time = time.time()

    ####################### MODFIED SECTION ##########################

    train_indices = [i[0] for i in train_test]
    test_splits = {}

    # NOTE: ttc stands for train-test combinations
    for is_test_tuple, ttc_name in zip(train_test_combinations, train_test_names):
        # is_test_tuple ~= (0, 1, 1, 0)
        test_indices = [
            ax_train_test[is_test]
            for is_test, ax_train_test in zip(is_test_tuple, train_test)
        ]
        test_splits[ttc_name] = _multipartite_safe_split(
            estimator, X, y, test_indices, train_indices, pairwise=pairwise
        )

    X_train, y_train = _multipartite_safe_split(
        estimator, X, y, train_indices, pairwise=pairwise
    )
    result = {}
    try:
        if y_train is None:
            estimator.fit(X_train, **fit_params)
        else:
            estimator.fit(X_train, y_train, **fit_params)

    except Exception:
        # Note fit time as time until error
        fit_time = time.time() - start_time
        score_time = 0.0
        if error_score == "raise":
            raise
        elif isinstance(error_score, numbers.Number):
            if isinstance(scorer, dict):
                test_scores = {
                    ttc_name + "_scores": {name: error_score for name in scorer}
                    for ttc_name in test_splits.keys()
                }
                # NOTE: train_score is automatically included by
                # _check_train_test_combinations, if requested.
            else:
                test_scores = {
                    ttc_name + "_scores": error_score for ttc_name in test_splits.keys()
                }
        result["fit_error"] = format_exc()
    else:
        result["fit_error"] = None

        fit_time = time.time() - start_time
        test_scores = {}
        for ttc_name, (X_test, y_test) in test_splits.items():
            test_scores[ttc_name + "_scores"] = _score(
                estimator, X_test, y_test.reshape(-1), scorer, error_score
            )
        score_time = time.time() - start_time - fit_time

    if verbose > 1:
        total_time = score_time + fit_time
        end_msg = f"[CV{progress_msg}] END "
        result_msg = params_msg + (";" if params_msg else "")
        if verbose > 2:
            # if isinstance(test_scores, dict):
            for scorer_name in sorted(test_scores):
                result_msg += f" {scorer_name}: ("
                result_msg += f"test={test_scores})"  # FIXME
        result_msg += f" total time={logger.short_format_time(total_time)}"

        # Right align the result_msg
        end_msg += "." * (80 - len(end_msg) - len(result_msg))
        end_msg += result_msg
        print(end_msg)

    result.update(test_scores)

    if return_n_test_samples:
        for ttc_name, (X_test, _) in test_splits.items():
            result[f"n_{ttc_name}_samples"] = tuple(x.shape[0] for x in X_test)
    if return_times:
        result["fit_time"] = fit_time
        result["score_time"] = score_time
    if return_parameters:
        result["parameters"] = parameters
    if return_estimator:
        result["estimator"] = estimator
    return result


# NOTE: Originally in sklearn.utils.metaestimators
def _multipartite_safe_split(
    estimator,
    X,
    y,
    indices,
    train_indices=None,
    pairwise=False,  # possibility to override estimator._more_tags['pairwise']
):
    """Create subset of n-dimensional dataset.

    Slice X, y according to indices for n-dimensional cross-validation.

    .. versionchanged:: hypertree
        No ``estimator`` or ``train_indices`` parameters are necessary by now.

    Parameters
    ----------
    X : list-like of array-like, sparse matrix or iterable
        The feature matrices for each axis.

    y : array-like, sparse matrix or iterable
        Targets to be indexed.

    indices : iterator of array of int
        Slice to select from X and y.

    Returns
    -------
    X_subset : list of array-like, sparse matrix or list
        Indexed data.

    y_subset : array-like, sparse matrix or list
        Indexed targets.

    """
    n_dim = y.ndim
    # TODO: further checking in another function may be adeuqate.
    # X_subset = _safe_indexing(X, indices)
    if not (len(X) == len(indices) == n_dim):
        raise ValueError(
            "Incompatible dimensions. One must ensure "
            "len(X) == len(indices) == y.ndim"
        )

    if pairwise and not _safe_tags(estimator, key="pairwise"):
        warnings.warn(
            "Setting pairwise=True, overriding estimator's " "pairwise=False tag."
        )

    if pairwise or _safe_tags(estimator, key="pairwise"):
        X_subset = []
        for i in range(n_dim):
            Xi, ind = X[i], indices[i]
            if not hasattr(Xi, "shape"):
                raise ValueError(
                    "Precomputed kernels or affinity matrices have "
                    "to be passed as arrays or sparse matrices."
                )
            # Xi is a precomputed square kernel matrix
            if Xi.shape[0] != Xi.shape[1]:
                raise ValueError("Xi should be a square kernel matrix")
            if train_indices is None:
                X_subset.append(Xi[np.ix_(ind, ind)])
            else:
                X_subset.append(Xi[np.ix_(ind, train_indices[i])])
    else:
        X_subset = [Xax[i] for Xax, i in zip(X, indices)]

    if y is not None:
        y_subset = y[np.ix_(*indices)]
    else:
        y_subset = None

    return X_subset, y_subset


# NOTE: Originally in sklearn.utils.__init__
# FIXME: unused by now.
def _multipartite_safe_indexing(X, indices):
    """Return rows, items or columns of X using indices.

    .. warning::

        This utility is documented, but **private**. This means that
        backward compatibility might be broken without any deprecation
        cycle.

    .. versionchanged:: hypertree
        The original sklearn function has an additional argument to specifying
        the axis of indexing, with it being required to be 0 or 1. Since we are
        in the n-dimensional domain, this restriction is eliminated and the axis
        information must be implicit in the `indices`' format, such as
        (slice(None), [3, 4, 5]) to get the columns 3, 4 and 5 of an ndarray (e-
        quivalent to np.s_[:, [3, 4, 5]]).

    Parameters
    ----------
    X : array-like, sparse-matrix, list, pandas.DataFrame, pandas.Series
        Data from which to sample rows, items or columns. `list` are only
        supported when `axis=0`.
    indices : bool, int, str, slice, array-like

    Returns
    -------
    subset
        Subset of X on axis 0 or 1.

    Notes
    -----
    CSR, CSC, and LIL sparse matrices are supported. COO sparse matrices are
    not supported.
    """
    if indices is None:
        return X

    # TODO: use isinstance.
    # if hasattr(X, "iloc"):
    #     return _pandas_indexing(X, indices, indices_dtype, axis=axis)
    # elif hasattr(X, "shape"):
    #     return _array_indexing(X, indices, indices_dtype, axis=axis)
    if isinstance(X, list):
        return sklearn.utils._list_indexing(X, indices, indices_dtype)
    else:
        return X[indices]



from bipartite_learn import BipartiteDecisionTreeRegressor
from .utils.make_examples import make_interaction_blobs
from .utils.test_utils import stopwatch
from .test_bipartite_trees import assert_equal_leaves
from time import perf_counter


def main():
    result_gmo = []
    result_gso = []

    for n_samples, n_features in (
        ((5, 5), (50, 50)),
        ((10, 9), (50, 100)),
        ((100, 90), (50, 40)),
        ((100, 90), (50, 40)),
        ((10, 10), (100, 200)),
        ((10, 10), (500, 400)),
        ((10, 10), (5000, 4000)),
        ((10, 10), (50000, 40000)),
    ):
        X, Y = make_interaction_blobs(
            random_state=0,
            n_samples=n_samples,
            n_features=n_features,
            noise=2.0,
            centers=10,
        )

        print('Training GMO tree')
        t0 = perf_counter()
        tree_gmo = BipartiteDecisionTreeRegressor(
            criterion='squared_error_gso',
            bipartite_adapter='gmosa',
        ).fit(X, Y)
        result_gmo.append(perf_counter() - t0)

        print('Training GSO tree')
        t0 = perf_counter()
        tree_gso = BipartiteDecisionTreeRegressor(
            criterion='squared_error',
            bipartite_adapter='gso',
        ).fit(X, Y)
        result_gso.append(perf_counter() - t0)

        assert_equal_leaves(tree_gmo.tree_, tree_gso.tree_)

        print('GMO:', result_gmo)
        print('GSO:', result_gso)


if __name__ == '__main__':
    main()


import numpy as np
from abc import ABCMeta, abstractmethod, abstractproperty

from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.validation import (
    check_X_y, check_array, _check_y, _num_features,
)
from imblearn.base import SamplerMixin
from .utils import check_partiteness, _X_is_multipartite


class BaseMultipartiteEstimator(BaseEstimator):
    """Base class for multipartite estimators.
    """
    # _partiteness should be set to k for an estimator that accepts only
    # k-partite input (in fit(), k = len(X) = y.ndim). None implies that any
    # any k >= 2 is accepted.
    _partiteness = None

    def score(self, X, y, sample_weight=None):
        # TODO: multipartite multi-output (y.ndim = 3).
        return super().score(X, y.reshape(-1), sample_weight)

    def _more_tags(self):
        # Determines that it can receive X as a list of Xs, one for each y axis,
        # that is, a list of objects of any type in self._get_tags()["X_types"]
        return dict(multipartite=True)

    def _validate_data(
        self,
        X="no_validation",
        y="no_validation",
        reset=True,
        validate_separately=False,
        **check_params,
    ):
        for Xax in X:
            self._check_feature_names(Xax, reset=reset)

        if y is None and self._get_tags()["requires_y"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator "
                "requires y to be passed, but the target y is None."
            )

        no_val_X = isinstance(X, str) and X == "no_validation"
        no_val_y = y is None or isinstance(y, str) and y == "no_validation"

        default_check_params = {"estimator": self}
        check_params = {**default_check_params, **check_params}

        if no_val_X and no_val_y:
            raise ValueError("Validation should be done on X, y or both.")
        elif not no_val_X and no_val_y:
            # TODO: better way of deciding. We still accept nd_arrays in
            #       predict, considering them as molten multipartite X
            #       (see docs for :module:melter)
            if _X_is_multipartite(X):
                check_partiteness(X, estimator=self)
                for ax in range(len(X)):
                    X[ax] = check_array(X[ax], input_name="X", **check_params)
            else:
                X = check_array(X, input_name="X", **check_params)
            out = X
        elif no_val_X and not no_val_y:
            check_partiteness(y=y, estimator=self)
            y = _check_y(y, **check_params)
            out = y
        else:
            check_partiteness(X, y, estimator=self)
            if len(X) != y.ndim:
                raise ValueError("Incompatible X and y given. The number of "
                                 "attribute matrices (len(X)) must correspond "
                                 f"to y's number of dimensions, but {len(X)=} "
                                 f"and {y.ndim=}.")
            if validate_separately:
                # We need this because some estimators validate X and y
                # separately, and in general, separately calling check_array()
                # on X and y isn't equivalent to just calling check_X_y()
                # :(
                check_X_params, check_y_params = validate_separately
                if "estimator" not in check_X_params:
                    check_X_params = {**default_check_params, **check_X_params}

                for ax in range(len(X)):
                    X[ax] = check_array(X[ax], input_name="X", **check_X_params)

                if "estimator" not in check_y_params:
                    check_y_params = {**default_check_params, **check_y_params}
                y = check_array(y, input_name="y", **check_y_params)
            else:
                X[0], y = check_X_y(X[0], y, multi_output=True, **check_params)
                _y = y
                for ax in range(1, len(X)):
                    _y = np.moveaxis(_y, 0, -1)
                    X[ax], _ = check_X_y(X[ax], _y, multi_output=True,
                                         **check_params)
            out = X, y

        if not no_val_X and check_params.get("ensure_2d", True):
            self._check_n_features(X, reset=reset)

        return out

    def _check_n_features(self, X, reset):
        """Set the `n_features_in_` attribute, or check against it.
        Parameters
        ----------
        X : list of {ndarray, sparse matrix} of shapes (n_samples[i], \
        n_features[i])
            The input samples.
        reset : bool
            If True, the `n_features_in_` attribute is set to
            `sum(Xi.shape[1] for Xi in X)`.
            If False and the attribute exists, then check that it is equal to
            `X.shape[1]`. If False and the attribute does *not* exist, then
            the check is skipped.
            .. note::
               It is recommended to call reset=True in `fit` and in the first
               call to `partial_fit`. All other methods that validate `X`
               should set `reset=False`.
        """
        if not _X_is_multipartite(X):  # Catches molten X (see ../melter.py)
            return super()._check_n_features(X, reset=reset)
        try:
            axes_n_features = tuple(_num_features(Xi) for Xi in X)
            n_features = sum(axes_n_features)
        except TypeError as e:
            if not reset and hasattr(self, "n_features_in_"):
                raise ValueError(
                    "X does not contain any features, but "
                    f"{self.__class__.__name__} is expecting "
                    f"{self.n_features_in_} features"
                ) from e
            # If the number of features is not defined and reset=True,
            # then we skip this check
            return

        if reset:
            self.n_features_in_ = n_features
            self.axes_n_features_in_ = axes_n_features
            return

        if not hasattr(self, "n_features_in_"):
            # Skip this check if the expected number of expected input features
            # was not recorded by calling fit first. This is typically the case
            # for stateless transformers.
            return

        if n_features != self.n_features_in_:
            raise ValueError(
                f"X has {n_features} features, but {self.__class__.__name__} "
                f"is expecting {self.n_features_in_} features as input."
            )


class BaseBipartiteEstimator(BaseMultipartiteEstimator):
    _partiteness = 2


class MultipartiteTransformerMixin(TransformerMixin):
    """Mixin for multipartite transformers."""


# FIXME: make it consistent with imblearn hierarchy
class MultipartiteSamplerMixin(BaseMultipartiteEstimator, SamplerMixin):
    pass


class BaseMultipartiteSampler(MultipartiteSamplerMixin):
    """Base class for multipartite samplers.
    """
    sampling_strategy = "auto"
    _sampling_type = "clean-sampling"

    def fit_resample(self, X, y):
        # TODO: do not bypass input validation.
        return self._fit_resample(X, y)

    def _check_X_y(self, X, y, accept_sparse=None):
        if accept_sparse is None:
            accept_sparse = ["csr", "csc"]
        # y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        X, y = self._validate_data(X, y, reset=True, accept_sparse=accept_sparse)
        return X, y, None

    def _more_tags(self):
        return {"X_types": ["n_partite", "2darray", "sparse", "dataframe"]}